import asyncio
import json
import logging
import re
import time
from typing import Dict, List, cast

import aiohttp
from jsonpath import jsonpath
from tenacity import retry, retry_if_exception_type, stop_after_delay, wait_exponential

from .api import AppApi, WebApi
from .exceptions import (
    LiveRoomEncrypted,
    LiveRoomHidden,
    LiveRoomLocked,
    NoAlternativeStreamAvailable,
    NoStreamAvailable,
    NoStreamCodecAvailable,
    NoStreamFormatAvailable,
    NoStreamQualityAvailable,
)
from .models import LiveStatus, RoomInfo, UserInfo
from .typing import ApiPlatform, QualityNumber, ResponseData, StreamCodec, StreamFormat

__all__ = ('Live',)

logger = logging.getLogger(__name__)

_INFO_PATTERN = re.compile(
    rb'<script>\s*window\.__NEPTUNE_IS_MY_WAIFU__\s*=\s*(\{.*?\})\s*</script>'
)
_LIVE_STATUS_PATTERN = re.compile(rb'"live_status"\s*:\s*(\d)')


class Live:
    def __init__(self, room_id: int, user_agent: str = '', cookie: str = '') -> None:
        self._room_id = room_id
        self._user_agent = user_agent
        self._cookie = cookie
        self._html_page_url = f'https://live.bilibili.com/{room_id}'

        self._room_info: RoomInfo
        self._user_info: UserInfo

    @property
    def user_agent(self) -> str:
        return self._user_agent

    @user_agent.setter
    def user_agent(self, value: str) -> None:
        self._user_agent = value

    @property
    def cookie(self) -> str:
        return self._cookie

    @cookie.setter
    def cookie(self, value: str) -> None:
        self._cookie = value

    @property
    def headers(self) -> Dict[str, str]:
        return {
            'Referer': 'https://live.bilibili.com/',
            'Connection': 'Keep-Alive',
            'Accept-Encoding': 'gzip',
            'User-Agent': self._user_agent,
            'Cookie': self._cookie,
        }

    @property
    def session(self) -> aiohttp.ClientSession:
        return self._session

    @property
    def appapi(self) -> AppApi:
        return self._appapi

    @property
    def webapi(self) -> WebApi:
        return self._webapi

    @property
    def room_id(self) -> int:
        return self._room_id

    @property
    def room_info(self) -> RoomInfo:
        return self._room_info

    @property
    def user_info(self) -> UserInfo:
        return self._user_info

    async def init(self) -> None:
        self._session = aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(limit=200),
            headers=self.headers,
            raise_for_status=True,
            trust_env=True,
        )
        self._appapi = AppApi(self._session)
        self._webapi = WebApi(self._session)

        self._room_info = await self.get_room_info()
        self._user_info = await self.get_user_info(self._room_info.uid)

    async def deinit(self) -> None:
        await self._session.close()

    async def get_live_status(self) -> LiveStatus:
        try:
            # frequent requests will be intercepted by the server's firewall!
            live_status = await self._get_live_status_via_api()
        except Exception:
            # more cpu consumption
            live_status = await self._get_live_status_via_html_page()

        return LiveStatus(live_status)

    def is_living(self) -> bool:
        return self._room_info.live_status == LiveStatus.LIVE

    async def check_connectivity(self) -> bool:
        try:
            await self._session.head('https://live.bilibili.com/', timeout=3)
        except (aiohttp.ClientConnectionError, asyncio.TimeoutError):
            return False
        else:
            return True

    async def update_info(self, raise_exception: bool = False) -> bool:
        return all(
            await asyncio.gather(
                self.update_user_info(raise_exception=raise_exception),
                self.update_room_info(raise_exception=raise_exception),
            )
        )

    async def update_user_info(self, raise_exception: bool = False) -> bool:
        try:
            self._user_info = await self.get_user_info(self._room_info.uid)
        except Exception as e:
            logger.error(f'Failed to update user info: {repr(e)}')
            if raise_exception:
                raise
            return False
        else:
            return True

    async def update_room_info(self, raise_exception: bool = False) -> bool:
        try:
            self._room_info = await self.get_room_info()
        except Exception as e:
            logger.error(f'Failed to update room info: {repr(e)}')
            if raise_exception:
                raise
            return False
        else:
            return True

    @retry(
        retry=retry_if_exception_type(
            (asyncio.TimeoutError, aiohttp.ClientError, ValueError)
        ),
        wait=wait_exponential(max=10),
        stop=stop_after_delay(60),
    )
    async def get_room_info(self) -> RoomInfo:
        try:
            # frequent requests will be intercepted by the server's firewall!
            room_info_data = await self._get_room_info_via_api()
        except Exception:
            # more cpu consumption
            room_info_data = await self._get_room_info_via_html_page()
        return RoomInfo.from_data(room_info_data)

    @retry(
        retry=retry_if_exception_type((asyncio.TimeoutError, aiohttp.ClientError)),
        wait=wait_exponential(max=10),
        stop=stop_after_delay(60),
    )
    async def get_user_info(self, uid: int) -> UserInfo:
        try:
            user_info_data = await self._webapi.get_user_info(uid)
            return UserInfo.from_web_api_data(user_info_data)
        except Exception:
            user_info_data = await self._appapi.get_user_info(uid)
            return UserInfo.from_app_api_data(user_info_data)

    async def get_timestamp(self) -> int:
        try:
            ts = await self.get_server_timestamp()
        except Exception as e:
            logger.warning(f'Failed to get timestamp from server: {repr(e)}')
            ts = int(time.time())
        return ts

    async def get_server_timestamp(self) -> int:
        # the timestamp on the server at the moment in seconds
        return await self._webapi.get_timestamp()

    async def get_live_stream_url(
        self,
        qn: QualityNumber = 10000,
        *,
        api_platform: ApiPlatform = 'web',
        stream_format: StreamFormat = 'flv',
        stream_codec: StreamCodec = 'avc',
        select_alternative: bool = False,
    ) -> str:
        if api_platform == 'web':
            info = await self._webapi.get_room_play_info(self._room_id, qn)
        else:
            info = await self._appapi.get_room_play_info(self._room_id, qn)

        self._check_room_play_info(info)

        streams = jsonpath(info, '$.playurl_info.playurl.stream[*]')
        if not streams:
            raise NoStreamAvailable(stream_format, stream_codec, qn)
        formats = jsonpath(
            streams, f'$[*].format[?(@.format_name == "{stream_format}")]'
        )
        if not formats:
            raise NoStreamFormatAvailable(stream_format, stream_codec, qn)
        codecs = jsonpath(formats, f'$[*].codec[?(@.codec_name == "{stream_codec}")]')
        if not codecs:
            raise NoStreamCodecAvailable(stream_format, stream_codec, qn)
        codec = codecs[0]

        accept_qn = cast(List[QualityNumber], codec['accept_qn'])
        if qn not in accept_qn or codec['current_qn'] != qn:
            raise NoStreamQualityAvailable(stream_format, stream_codec, qn)

        urls = [i['host'] + codec['base_url'] + i['extra'] for i in codec['url_info']]

        if not select_alternative:
            return urls[0]

        try:
            return urls[1]
        except IndexError:
            raise NoAlternativeStreamAvailable(stream_format, stream_codec, qn)

    def _check_room_play_info(self, data: ResponseData) -> None:
        if data.get('is_hidden'):
            raise LiveRoomHidden()
        if data.get('is_locked'):
            raise LiveRoomLocked()
        if data.get('encrypted') and not data.get('pwd_verified'):
            raise LiveRoomEncrypted()

    async def _get_live_status_via_api(self) -> int:
        room_info_data = await self._get_room_info_via_api()
        return int(room_info_data['live_status'])

    async def _get_room_info_via_api(self) -> ResponseData:
        try:
            info_data = await self._webapi.get_info_by_room(self._room_id)
            room_info_data = info_data['room_info']
        except Exception:
            try:
                info_data = await self._appapi.get_info_by_room(self._room_id)
                room_info_data = info_data['room_info']
            except Exception:
                room_info_data = await self._webapi.get_info(self._room_id)

        return room_info_data

    async def _get_live_status_via_html_page(self) -> int:
        async with self._session.get(self._html_page_url) as response:
            data = await response.read()

        m = _LIVE_STATUS_PATTERN.search(data)
        assert m is not None, data

        return int(m.group(1))

    async def _get_room_info_via_html_page(self) -> ResponseData:
        info_res = await self._get_room_info_res_via_html_page()
        return info_res['room_info']

    async def _get_room_play_info_via_html_page(self) -> ResponseData:
        return await self._get_room_init_res_via_html_page()

    async def _get_room_info_res_via_html_page(self) -> ResponseData:
        info = await self._get_info_via_html_page()
        if info['roomInfoRes']['code'] != 0:
            raise ValueError(f"Invaild roomInfoRes: {info['roomInfoRes']}")
        return info['roomInfoRes']['data']

    async def _get_room_init_res_via_html_page(self) -> ResponseData:
        info = await self._get_info_via_html_page()
        if info['roomInitRes']['code'] != 0:
            raise ValueError(f"Invaild roomInitRes: {info['roomInitRes']}")
        return info['roomInitRes']['data']

    async def _get_info_via_html_page(self) -> ResponseData:
        async with self._session.get(self._html_page_url) as response:
            data = await response.read()

        match = _INFO_PATTERN.search(data)
        if not match:
            raise ValueError('Can not extract info from html page')

        string = match.group(1).decode(encoding='utf8')
        return json.loads(string)



from __future__ import annotations

import logging
from typing import Optional
from urllib.parse import urlparse

import requests
import urllib3
from reactivex import Observable, abc
from reactivex import operators as ops

from ...bili.exceptions import (
    LiveRoomEncrypted,
    LiveRoomHidden,
    LiveRoomLocked,
    NoAlternativeStreamAvailable,
    NoStreamAvailable,
    NoStreamCodecAvailable,
    NoStreamFormatAvailable,
    NoStreamQualityAvailable,
)
from ...bili.live import Live
from ...utils import operators as utils_ops
from ...utils.mixins import AsyncCooperationMixin
from ..stream_param_holder import StreamParamHolder, StreamParams

__all__ = ('StreamURLResolver',)


logger = logging.getLogger(__name__)
logging.getLogger(urllib3.__name__).setLevel(logging.WARNING)


class StreamURLResolver(AsyncCooperationMixin):
    def __init__(self, live: Live, stream_param_holder: StreamParamHolder) -> None:
        super().__init__()
        self._live = live
        self._stream_param_holder = stream_param_holder
        self._stream_url: str = ''
        self._stream_host: str = ''
        self._stream_params: Optional[StreamParams] = None

    @property
    def stream_url(self) -> str:
        return self._stream_url

    @property
    def stream_host(self) -> str:
        return self._stream_host

    def _reset(self) -> None:
        self._stream_url = ''
        self._stream_host = ''
        self._stream_params = None

    def __call__(self, source: Observable[StreamParams]) -> Observable[str]:
        self._reset()
        return self._solve(source).pipe(
            ops.do_action(on_error=self._before_retry),
            utils_ops.retry(delay=1, should_retry=self._should_retry),
        )

    def _solve(self, source: Observable[StreamParams]) -> Observable[str]:
        def subscribe(
            observer: abc.ObserverBase[str],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            def on_next(params: StreamParams) -> None:
                if self._can_resue_url(params):
                    observer.on_next(self._stream_url)
                    return

                try:
                    logger.info(
                        f'Getting the live stream url... '
                        f'qn: {params.quality_number}, '
                        f'format: {params.stream_format}, '
                        f'api platform: {params.api_platform}, '
                        f'use alternative stream: {params.use_alternative_stream}'
                    )
                    url = self._run_coroutine(
                        self._live.get_live_stream_url(
                            params.quality_number,
                            api_platform=params.api_platform,
                            stream_format=params.stream_format,
                            select_alternative=params.use_alternative_stream,
                        )
                    )
                except Exception as e:
                    logger.warning(f'Failed to get live stream url: {repr(e)}')
                    observer.on_error(e)
                else:
                    logger.info(f"Got live stream url: '{url}'")
                    self._stream_url = url
                    self._stream_host = urlparse(url).hostname or ''
                    self._stream_params = params
                    observer.on_next(url)

            return source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

        return Observable(subscribe)

    def _can_resue_url(self, params: StreamParams) -> bool:
        if params == self._stream_params and self._stream_url:
            try:
                response = requests.get(
                    self._stream_url, stream=True, headers=self._live.headers
                )
                response.raise_for_status()
            except Exception:
                return False
            else:
                return True
        else:
            return False

    def _should_retry(self, exc: Exception) -> bool:
        if isinstance(
            exc,
            (
                NoStreamAvailable,
                NoStreamCodecAvailable,
                NoStreamFormatAvailable,
                NoStreamQualityAvailable,
                NoAlternativeStreamAvailable,
            ),
        ):
            return True
        else:
            return False

    def _before_retry(self, exc: Exception) -> None:
        try:
            raise exc
        except (NoStreamAvailable, NoStreamCodecAvailable, NoStreamFormatAvailable):
            pass
        except NoStreamQualityAvailable:
            qn = self._stream_param_holder.quality_number
            logger.info(
                f'The specified stream quality ({qn}) is not available, '
                'will using the original stream quality (10000) instead.'
            )
            self._stream_param_holder.fall_back_quality()
        except NoAlternativeStreamAvailable:
            logger.debug(
                'No alternative stream url available, '
                'will using the primary stream url instead.'
            )
            self._stream_param_holder.use_alternative_stream = False
            self._stream_param_holder.rotate_api_platform()
        except LiveRoomHidden:
            pass
        except LiveRoomLocked:
            pass
        except LiveRoomEncrypted:
            pass
        except Exception:
            pass



import asyncio
import logging
import shutil
from contextlib import suppress

from ..event.event_emitter import EventEmitter, EventListener
from ..exception import exception_callback
from ..logging.room_id import aio_task_with_room_id
from ..utils.mixins import AsyncStoppableMixin, SwitchableMixin
from .helpers import is_space_enough
from .models import DiskUsage

__all__ = 'SpaceMonitor', 'SpaceEventListener'


logger = logging.getLogger(__name__)


class SpaceEventListener(EventListener):
    async def on_space_no_enough(
        self, path: str, threshold: int, disk_usage: DiskUsage
    ) -> None:
        ...


class SpaceMonitor(
    EventEmitter[SpaceEventListener], SwitchableMixin, AsyncStoppableMixin
):
    def __init__(
        self,
        path: str,
        *,
        check_interval: int = 60,  # seconds
        space_threshold: int = 1073741824,  # 1GB
    ) -> None:
        super().__init__()
        self.path = path
        self.check_interval = check_interval
        self.space_threshold = space_threshold
        self._monitoring: bool = False

    def _do_enable(self) -> None:
        asyncio.create_task(self.start())
        logger.debug('Enabled space monitor')

    def _do_disable(self) -> None:
        asyncio.create_task(self.stop())
        logger.debug('Disabled space monitor')

    async def _do_start(self) -> None:
        self._create_polling_task()

    async def _do_stop(self) -> None:
        await self._cancel_polling_task()

    def _create_polling_task(self) -> None:
        self._polling_task = asyncio.create_task(self._polling_loop())
        self._polling_task.add_done_callback(exception_callback)

    async def _cancel_polling_task(self) -> None:
        self._polling_task.cancel()
        with suppress(asyncio.CancelledError):
            await self._polling_task

    @aio_task_with_room_id
    async def _polling_loop(self) -> None:
        while True:
            if not is_space_enough(self.path, self.space_threshold):
                logger.warning('No enough disk space left')
                await self._emit_space_no_enough()
            await asyncio.sleep(self.check_interval)

    async def _emit_space_no_enough(self) -> None:
        usage = DiskUsage(*shutil.disk_usage(self.path))
        await self._emit('space_no_enough', self.path, self.space_threshold, usage)



from .helpers import shadow_settings, update_settings
from .models import (
    DEFAULT_SETTINGS_FILE,
    DanmakuOptions,
    DanmakuSettings,
    EmailMessageTemplateSettings,
    EmailNotificationSettings,
    EmailSettings,
    EnvSettings,
    HeaderOptions,
    HeaderSettings,
    LoggingSettings,
    NotificationSettings,
    NotifierSettings,
    OutputSettings,
    PostprocessingOptions,
    PostprocessingSettings,
    PushdeerMessageTemplateSettings,
    PushdeerNotificationSettings,
    PushdeerSettings,
    PushplusMessageTemplateSettings,
    PushplusNotificationSettings,
    PushplusSettings,
    RecorderOptions,
    RecorderSettings,
    ServerchanMessageTemplateSettings,
    ServerchanNotificationSettings,
    ServerchanSettings,
    Settings,
    SettingsIn,
    SettingsOut,
    SpaceSettings,
    TaskOptions,
    TaskSettings,
    TelegramMessageTemplateSettings,
    TelegramNotificationSettings,
    TelegramSettings,
    WebHookSettings,
)
from .setting_manager import SettingsManager

__all__ = (
    'DEFAULT_SETTINGS_FILE',
    'EnvSettings',
    'Settings',
    'SettingsIn',
    'SettingsOut',
    'HeaderOptions',
    'HeaderSettings',
    'DanmakuOptions',
    'DanmakuSettings',
    'RecorderOptions',
    'RecorderSettings',
    'PostprocessingSettings',
    'PostprocessingOptions',
    'TaskOptions',
    'TaskSettings',
    'OutputSettings',
    'LoggingSettings',
    'SpaceSettings',
    'EmailMessageTemplateSettings',
    'ServerchanMessageTemplateSettings',
    'PushdeerMessageTemplateSettings',
    'PushplusMessageTemplateSettings',
    'TelegramMessageTemplateSettings',
    'EmailSettings',
    'ServerchanSettings',
    'PushdeerSettings',
    'PushplusSettings',
    'TelegramSettings',
    'NotifierSettings',
    'NotificationSettings',
    'EmailNotificationSettings',
    'ServerchanNotificationSettings',
    'PushdeerNotificationSettings',
    'PushplusNotificationSettings',
    'TelegramNotificationSettings',
    'WebHookSettings',
    'update_settings',
    'shadow_settings',
    'SettingsManager',
)



from functools import partial
from typing import Callable, Iterable, Iterator, List, Optional, cast

from fastapi import Query

from blrec.setting.typing import KeyOfSettings, KeySetOfSettings

from .schemas import DataSelection, AliasKeyOfSettings
from ..bili.models import LiveStatus
from ..task import RunningStatus, TaskData
from ..utils.string import snake_case


TaskDataFilter = Callable[[Iterable[TaskData]], Iterator[TaskData]]


def task_data_filter(
    select: DataSelection = Query('all'),
) -> TaskDataFilter:
    if select == DataSelection.ALL:
        def func(data: TaskData) -> bool:
            return True
    # live status
    elif select == DataSelection.PREPARING:
        def func(data: TaskData) -> bool:
            return data.room_info.live_status == LiveStatus.PREPARING
    elif select == DataSelection.LIVING:
        def func(data: TaskData) -> bool:
            return data.room_info.live_status == LiveStatus.LIVE
    elif select == DataSelection.ROUNDING:
        def func(data: TaskData) -> bool:
            return data.room_info.live_status == LiveStatus.ROUND
    # task status
    elif select == DataSelection.MONITOR_ENABLED:
        def func(data: TaskData) -> bool:
            return data.task_status.monitor_enabled
    elif select == DataSelection.MONITOR_DISABLED:
        def func(data: TaskData) -> bool:
            return not data.task_status.monitor_enabled
    elif select == DataSelection.RECORDER_ENABLED:
        def func(data: TaskData) -> bool:
            return data.task_status.recorder_enabled
    elif select == DataSelection.RECORDER_DISABLED:
        def func(data: TaskData) -> bool:
            return not data.task_status.recorder_enabled
    # task running status
    elif select == DataSelection.STOPPED:
        def func(data: TaskData) -> bool:
            return data.task_status.running_status == RunningStatus.STOPPED
    elif select == DataSelection.WAITTING:
        def func(data: TaskData) -> bool:
            return data.task_status.running_status == RunningStatus.WAITING
    elif select == DataSelection.RECORDING:
        def func(data: TaskData) -> bool:
            return data.task_status.running_status == RunningStatus.RECORDING
    elif select == DataSelection.REMUXING:
        def func(data: TaskData) -> bool:
            return data.task_status.running_status == RunningStatus.REMUXING

    return partial(filter, func)  # type: ignore


def settings_include_set(
    include: Optional[List[AliasKeyOfSettings]] = Query(None),
) -> Optional[KeySetOfSettings]:
    return frozenset(map(key_of_settings, include)) if include else None


def settings_exclude_set(
    exclude: Optional[List[AliasKeyOfSettings]] = Query(None),
) -> Optional[KeySetOfSettings]:
    return frozenset(map(key_of_settings, exclude)) if exclude else None


def key_of_settings(key: AliasKeyOfSettings) -> KeyOfSettings:
    return cast(KeyOfSettings, snake_case(key))



from typing import Optional

import aiohttp


from .api import PypiApi
from .typing import Metadata


__all__ = (
    'get_project_metadata',
    'get_release_metadata',
    'get_latest_version_string',
)


async def get_project_metadata(project_name: str) -> Optional[Metadata]:
    async with aiohttp.ClientSession(raise_for_status=True) as session:
        api = PypiApi(session)
        return await api.get_project_metadata(project_name)


async def get_release_metadata(
    project_name: str, version: str
) -> Optional[Metadata]:
    async with aiohttp.ClientSession(raise_for_status=True) as session:
        api = PypiApi(session)
        return await api.get_release_metadata(project_name, version)


async def get_latest_version_string(project_name: str) -> Optional[str]:
    metadata = await get_project_metadata(project_name)
    if not metadata:
        return None
    return metadata['info']['version']



import json
from typing import Any, Dict, Iterable


from .io import FlvReader
from .common import (
    is_audio_tag, is_metadata_tag, is_script_tag, is_video_tag, read_tags,
    parse_metadata, find_metadata_tag,
)
from .operators import JoinPoint
from .utils import format_timestamp
from ..path import extra_metadata_path


def get_metadata(path: str) -> Dict[str, Any]:
    with open(path, mode='rb') as file:
        reader = FlvReader(file)
        reader.read_header()
        if (tag := find_metadata_tag(reversed(list(read_tags(reader, 5))))):
            return parse_metadata(tag)
        raise EOFError


def get_extra_metadata(flv_path: str) -> Dict[str, Any]:
    path = extra_metadata_path(flv_path)
    with open(path, 'rt', encoding='utf8') as file:
        return json.load(file)


def make_comment_for_joinpoints(join_points: Iterable[JoinPoint]) -> str:
    return (
        '流中断拼接详情\n' +
        '\n'.join((
            '时间戳：{}, 无缝拼接：{}'.format(
                format_timestamp(int(p.timestamp)),
                '是' if p.seamless else '否',
            )
            for p in join_points
        ))
    )


def is_valid_flv_file(path: str) -> bool:
    with open(path, mode='rb') as file:
        reader = FlvReader(file)

        try:
            flv_header = reader.read_header()
        except EOFError:
            return False

        MAX_TAGS_TO_FIND = 50
        has_metadata_tag = False
        has_audio_header_tag = False
        has_audio_data_tag = False
        has_video_header_tag = False
        has_video_data_tag = False

        for tag in read_tags(reader, MAX_TAGS_TO_FIND, no_body=True):
            if is_script_tag(tag):
                tag = tag.evolve(body=reader.read_body(tag))
                if is_metadata_tag(tag):
                    has_metadata_tag = True
            elif is_audio_tag(tag):
                if tag.is_aac_header():
                    has_audio_header_tag = True
                elif tag.is_aac_raw():
                    has_audio_data_tag = True
            elif is_video_tag(tag):
                if tag.is_avc_header():
                    has_video_header_tag = True
                elif tag.is_avc_nalu():
                    has_video_data_tag = True
            else:
                pass

            if (
                has_metadata_tag and (
                    flv_header.has_video() ==
                    has_video_header_tag ==
                    has_video_data_tag
                ) and (
                    flv_header.has_audio() ==
                    has_audio_header_tag ==
                    has_audio_data_tag
                )
            ):
                return True

        return False



from reactivex import Observable, Subject

from ..utils.patterns import Singleton
from .typing import Event

__all__ = ('EventCenter',)


class EventCenter(Singleton):
    def __init__(self) -> None:
        super().__init__()
        self._source: Subject[Event] = Subject()

    @property
    def events(self) -> Observable[Event]:
        return self._source

    def submit(self, event: Event) -> None:
        self._source.on_next(event)



from __future__ import annotations

import io
import logging
from typing import Final, List, Optional, Union

import urllib3
from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable

from ...bili.live import Live
from ...utils.io import wait_for
from ..stream_remuxer import StreamRemuxer
from .segment_fetcher import InitSectionData, SegmentData

__all__ = ('SegmentRemuxer',)


logger = logging.getLogger(__name__)
logging.getLogger(urllib3.__name__).setLevel(logging.WARNING)


class SegmentRemuxer:
    _MAX_SEGMENT_DATA_CACHE: Final = 3

    def __init__(self, live: Live) -> None:
        self._live = live
        self._timeout: float = 10
        self._stream_remuxer = StreamRemuxer(live.room_id, remove_filler_data=True)

    def __call__(
        self, source: Observable[Union[InitSectionData, SegmentData]]
    ) -> Observable[io.RawIOBase]:
        return self._remux(source)

    def _remux(
        self, source: Observable[Union[InitSectionData, SegmentData]]
    ) -> Observable[io.RawIOBase]:
        def subscribe(
            observer: abc.ObserverBase[io.RawIOBase],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()

            init_section_data: Optional[bytes] = None
            segment_data_cache: List[bytes] = []
            self._stream_remuxer.stop()

            def write(data: bytes) -> int:
                return wait_for(
                    self._stream_remuxer.input.write,
                    args=(data,),
                    timeout=self._timeout,
                )

            def on_next(data: Union[InitSectionData, SegmentData]) -> None:
                nonlocal init_section_data

                if isinstance(data, InitSectionData):
                    init_section_data = data.payload
                    segment_data_cache.clear()
                    self._stream_remuxer.stop()

                try:
                    if self._stream_remuxer.stopped:
                        self._stream_remuxer.start()
                        while True:
                            ready = self._stream_remuxer.wait(timeout=1)
                            if disposed:
                                return
                            if ready:
                                break

                        observer.on_next(RemuxedStream(self._stream_remuxer))

                        if init_section_data:
                            write(init_section_data)
                        if segment_data_cache:
                            for cached_data in segment_data_cache:
                                write(cached_data)
                        if isinstance(data, InitSectionData):
                            return

                    write(data.payload)
                except Exception as e:
                    logger.warning(f'Failed to write data to stream remuxer: {repr(e)}')
                    self._stream_remuxer.stop()

                segment_data_cache.append(data.payload)
                if len(segment_data_cache) > self._MAX_SEGMENT_DATA_CACHE:
                    segment_data_cache.pop(0)

            def dispose() -> None:
                nonlocal disposed
                disposed = True

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)


class RemuxedStream(io.RawIOBase):
    def __init__(
        self, stream_remuxer: StreamRemuxer, *, read_timeout: float = 10
    ) -> None:
        self._stream_remuxer = stream_remuxer
        self._read_timeout = read_timeout
        self._offset: int = 0

    def read(self, size: int = -1) -> bytes:
        if self._stream_remuxer.stopped:
            ready = self._stream_remuxer.wait(timeout=self._read_timeout)
            if not ready:
                logger.debug(
                    f'Stream remuxer not ready in {self._read_timeout} seconds'
                )
                raise EOFError

        try:
            data = wait_for(
                self._stream_remuxer.output.read,
                args=(size,),
                timeout=self._read_timeout,
            )
        except Exception as e:
            logger.warning(f'Failed to read data from stream remuxer: {repr(e)}')
            self._stream_remuxer.stop()
            raise EOFError
        else:
            assert data is not None
            self._offset += len(data)
            return data

    def tell(self) -> int:
        return self._offset

    def close(self) -> None:
        self._stream_remuxer.stop()



from __future__ import annotations

import asyncio
import time
import logging
from datetime import datetime
from typing import Iterator, Optional

import humanize

from ..bili.danmaku_client import DanmakuClient
from ..bili.live import Live
from ..bili.live_monitor import LiveEventListener, LiveMonitor
from ..bili.models import RoomInfo
from ..bili.typing import QualityNumber, StreamFormat
from ..event.event_emitter import EventEmitter, EventListener
from ..flv.operators import MetaData, StreamProfile
from ..logging.room_id import aio_task_with_room_id
from ..utils.mixins import AsyncStoppableMixin
from .cover_downloader import CoverDownloader, CoverSaveStrategy
from .danmaku_dumper import DanmakuDumper, DanmakuDumperEventListener
from .danmaku_receiver import DanmakuReceiver
from .raw_danmaku_dumper import RawDanmakuDumper, RawDanmakuDumperEventListener
from .raw_danmaku_receiver import RawDanmakuReceiver
from .stream_recorder import StreamRecorder, StreamRecorderEventListener

__all__ = 'RecorderEventListener', 'Recorder'


logger = logging.getLogger(__name__)


class RecorderEventListener(EventListener):
    async def on_recording_started(self, recorder: Recorder) -> None:
        ...

    async def on_recording_finished(self, recorder: Recorder) -> None:
        ...

    async def on_recording_cancelled(self, recorder: Recorder) -> None:
        ...

    async def on_video_file_created(self, recorder: Recorder, path: str) -> None:
        ...

    async def on_video_file_completed(self, recorder: Recorder, path: str) -> None:
        ...

    async def on_danmaku_file_created(self, recorder: Recorder, path: str) -> None:
        ...

    async def on_danmaku_file_completed(self, recorder: Recorder, path: str) -> None:
        ...

    async def on_raw_danmaku_file_created(self, recorder: Recorder, path: str) -> None:
        ...

    async def on_raw_danmaku_file_completed(
        self, recorder: Recorder, path: str
    ) -> None:
        ...


class Recorder(
    EventEmitter[RecorderEventListener],
    LiveEventListener,
    AsyncStoppableMixin,
    DanmakuDumperEventListener,
    RawDanmakuDumperEventListener,
    StreamRecorderEventListener,
):
    def __init__(
        self,
        live: Live,
        danmaku_client: DanmakuClient,
        live_monitor: LiveMonitor,
        out_dir: str,
        path_template: str,
        *,
        stream_format: StreamFormat = 'flv',
        quality_number: QualityNumber = 10000,
        fmp4_stream_timeout: int = 10,
        buffer_size: Optional[int] = None,
        read_timeout: Optional[int] = None,
        disconnection_timeout: Optional[int] = None,
        filesize_limit: int = 0,
        duration_limit: int = 0,
        danmu_uname: bool = False,
        record_gift_send: bool = False,
        record_free_gifts: bool = False,
        record_guard_buy: bool = False,
        record_super_chat: bool = False,
        save_cover: bool = False,
        cover_save_strategy: CoverSaveStrategy = CoverSaveStrategy.DEFAULT,
        save_raw_danmaku: bool = False,
    ) -> None:
        super().__init__()

        self._live = live
        self._danmaku_client = danmaku_client
        self._live_monitor = live_monitor
        self.save_raw_danmaku = save_raw_danmaku

        self._recording: bool = False
        self._stream_available: bool = False

        self._stream_recorder = StreamRecorder(
            self._live,
            out_dir=out_dir,
            path_template=path_template,
            stream_format=stream_format,
            quality_number=quality_number,
            fmp4_stream_timeout=fmp4_stream_timeout,
            buffer_size=buffer_size,
            read_timeout=read_timeout,
            disconnection_timeout=disconnection_timeout,
            filesize_limit=filesize_limit,
            duration_limit=duration_limit,
        )

        self._danmaku_receiver = DanmakuReceiver(danmaku_client)
        self._danmaku_dumper = DanmakuDumper(
            self._live,
            self._stream_recorder,
            self._danmaku_receiver,
            danmu_uname=danmu_uname,
            record_gift_send=record_gift_send,
            record_free_gifts=record_free_gifts,
            record_guard_buy=record_guard_buy,
            record_super_chat=record_super_chat,
        )
        self._raw_danmaku_receiver = RawDanmakuReceiver(danmaku_client)
        self._raw_danmaku_dumper = RawDanmakuDumper(
            self._live, self._stream_recorder, self._raw_danmaku_receiver
        )

        self._cover_downloader = CoverDownloader(
            self._live,
            self._stream_recorder,
            save_cover=save_cover,
            cover_save_strategy=cover_save_strategy,
        )

    @property
    def live(self) -> Live:
        return self._live

    @property
    def recording(self) -> bool:
        return self._recording

    @property
    def stream_format(self) -> StreamFormat:
        return self._stream_recorder.stream_format

    @stream_format.setter
    def stream_format(self, value: StreamFormat) -> None:
        self._stream_recorder.stream_format = value

    @property
    def quality_number(self) -> QualityNumber:
        return self._stream_recorder.quality_number

    @quality_number.setter
    def quality_number(self, value: QualityNumber) -> None:
        self._stream_recorder.quality_number = value

    @property
    def fmp4_stream_timeout(self) -> int:
        return self._stream_recorder.fmp4_stream_timeout

    @fmp4_stream_timeout.setter
    def fmp4_stream_timeout(self, value: int) -> None:
        self._stream_recorder.fmp4_stream_timeout = value

    @property
    def real_stream_format(self) -> Optional[StreamFormat]:
        return self._stream_recorder.real_stream_format

    @property
    def real_quality_number(self) -> Optional[QualityNumber]:
        return self._stream_recorder.real_quality_number

    @property
    def buffer_size(self) -> int:
        return self._stream_recorder.buffer_size

    @buffer_size.setter
    def buffer_size(self, value: int) -> None:
        self._stream_recorder.buffer_size = value

    @property
    def read_timeout(self) -> int:
        return self._stream_recorder.read_timeout

    @read_timeout.setter
    def read_timeout(self, value: int) -> None:
        self._stream_recorder.read_timeout = value

    @property
    def disconnection_timeout(self) -> int:
        return self._stream_recorder.disconnection_timeout

    @disconnection_timeout.setter
    def disconnection_timeout(self, value: int) -> None:
        self._stream_recorder.disconnection_timeout = value

    @property
    def danmu_uname(self) -> bool:
        return self._danmaku_dumper.danmu_uname

    @danmu_uname.setter
    def danmu_uname(self, value: bool) -> None:
        self._danmaku_dumper.danmu_uname = value

    @property
    def record_gift_send(self) -> bool:
        return self._danmaku_dumper.record_gift_send

    @record_gift_send.setter
    def record_gift_send(self, value: bool) -> None:
        self._danmaku_dumper.record_gift_send = value

    @property
    def record_free_gifts(self) -> bool:
        return self._danmaku_dumper.record_free_gifts

    @record_free_gifts.setter
    def record_free_gifts(self, value: bool) -> None:
        self._danmaku_dumper.record_free_gifts = value

    @property
    def record_guard_buy(self) -> bool:
        return self._danmaku_dumper.record_guard_buy

    @record_guard_buy.setter
    def record_guard_buy(self, value: bool) -> None:
        self._danmaku_dumper.record_guard_buy = value

    @property
    def record_super_chat(self) -> bool:
        return self._danmaku_dumper.record_super_chat

    @record_super_chat.setter
    def record_super_chat(self, value: bool) -> None:
        self._danmaku_dumper.record_super_chat = value

    @property
    def save_cover(self) -> bool:
        return self._cover_downloader.save_cover

    @save_cover.setter
    def save_cover(self, value: bool) -> None:
        self._cover_downloader.save_cover = value

    @property
    def cover_save_strategy(self) -> CoverSaveStrategy:
        return self._cover_downloader.cover_save_strategy

    @cover_save_strategy.setter
    def cover_save_strategy(self, value: CoverSaveStrategy) -> None:
        self._cover_downloader.cover_save_strategy = value

    @property
    def stream_url(self) -> str:
        return self._stream_recorder.stream_url

    @property
    def stream_host(self) -> str:
        return self._stream_recorder.stream_host

    @property
    def dl_total(self) -> int:
        return self._stream_recorder.dl_total

    @property
    def dl_rate(self) -> float:
        return self._stream_recorder.dl_rate

    @property
    def rec_elapsed(self) -> float:
        return self._stream_recorder.rec_elapsed

    @property
    def rec_total(self) -> int:
        return self._stream_recorder.rec_total

    @property
    def rec_rate(self) -> float:
        return self._stream_recorder.rec_rate

    @property
    def danmu_total(self) -> int:
        return self._danmaku_dumper.danmu_total

    @property
    def danmu_rate(self) -> float:
        return self._danmaku_dumper.danmu_rate

    @property
    def out_dir(self) -> str:
        return self._stream_recorder.out_dir

    @out_dir.setter
    def out_dir(self, value: str) -> None:
        self._stream_recorder.out_dir = value

    @property
    def path_template(self) -> str:
        return self._stream_recorder.path_template

    @path_template.setter
    def path_template(self, value: str) -> None:
        self._stream_recorder.path_template = value

    @property
    def filesize_limit(self) -> int:
        return self._stream_recorder.filesize_limit

    @filesize_limit.setter
    def filesize_limit(self, value: int) -> None:
        self._stream_recorder.filesize_limit = value

    @property
    def duration_limit(self) -> int:
        return self._stream_recorder.duration_limit

    @duration_limit.setter
    def duration_limit(self, value: int) -> None:
        self._stream_recorder.duration_limit = value

    @property
    def recording_path(self) -> Optional[str]:
        return self._stream_recorder.recording_path

    @property
    def metadata(self) -> Optional[MetaData]:
        return self._stream_recorder.metadata

    @property
    def stream_profile(self) -> StreamProfile:
        return self._stream_recorder.stream_profile

    def get_recording_files(self) -> Iterator[str]:
        if self._stream_recorder.recording_path is not None:
            yield self._stream_recorder.recording_path
        if self._danmaku_dumper.dumping_path is not None:
            yield self._danmaku_dumper.dumping_path

    def get_video_files(self) -> Iterator[str]:
        yield from self._stream_recorder.get_files()

    def get_danmaku_files(self) -> Iterator[str]:
        yield from self._danmaku_dumper.get_files()

    def can_cut_stream(self) -> bool:
        return self._stream_recorder.can_cut_stream()

    def cut_stream(self) -> bool:
        return self._stream_recorder.cut_stream()

    async def on_live_began(self, live: Live) -> None:
        logger.info('The live has began')
        self._print_live_info()
        await self._start_recording()

    async def on_live_ended(self, live: Live) -> None:
        logger.info('The live has ended')
        self._stream_available = False
        self._stream_recorder.stream_available_time = None
        await self._stop_recording()
        self._print_waiting_message()

    async def on_live_stream_available(self, live: Live) -> None:
        logger.debug('The live stream becomes available')
        self._stream_available = True
        self._stream_recorder.stream_available_time = await live.get_timestamp()
        await self._stream_recorder.start()

    async def on_live_stream_reset(self, live: Live) -> None:
        logger.warning('The live stream has been reset')
        if not self._recording:
            await self._start_recording()

    async def on_room_changed(self, room_info: RoomInfo) -> None:
        self._print_changed_room_info(room_info)
        self._stream_recorder.update_progress_bar_info()

    async def on_video_file_created(self, path: str, record_start_time: int) -> None:
        await self._emit('video_file_created', self, path)

    async def on_video_file_completed(self, path: str) -> None:
        await self._emit('video_file_completed', self, path)

    async def on_danmaku_file_created(self, path: str) -> None:
        await self._emit('danmaku_file_created', self, path)

    async def on_danmaku_file_completed(self, path: str) -> None:
        await self._emit('danmaku_file_completed', self, path)

    async def on_raw_danmaku_file_created(self, path: str) -> None:
        await self._emit('raw_danmaku_file_created', self, path)

    async def on_raw_danmaku_file_completed(self, path: str) -> None:
        await self._emit('raw_danmaku_file_completed', self, path)

    async def on_stream_recording_completed(self) -> None:
        logger.debug('Stream recording completed')
        await self._stop_recording()

    async def _do_start(self) -> None:
        self._live_monitor.add_listener(self)
        self._danmaku_dumper.add_listener(self)
        self._raw_danmaku_dumper.add_listener(self)
        self._stream_recorder.add_listener(self)
        logger.debug('Started recorder')

        self._print_live_info()
        if self._live.is_living():
            self._stream_available = True
            await self._start_recording()
        else:
            self._print_waiting_message()

    async def _do_stop(self) -> None:
        await self._stop_recording()
        self._live_monitor.remove_listener(self)
        self._danmaku_dumper.remove_listener(self)
        self._raw_danmaku_dumper.remove_listener(self)
        self._stream_recorder.remove_listener(self)
        logger.debug('Stopped recorder')

    async def _start_recording(self) -> None:
        if self._recording:
            return
        self._recording = True

        if self.save_raw_danmaku:
            self._raw_danmaku_dumper.enable()
            self._raw_danmaku_receiver.start()
        self._danmaku_dumper.enable()
        self._danmaku_receiver.start()
        self._cover_downloader.enable()

        await self._prepare()
        if self._stream_available:
            await self._stream_recorder.start()
        else:
            asyncio.create_task(self._guard())

        logger.info('Started recording')
        await self._emit('recording_started', self)

    async def _stop_recording(self) -> None:
        if not self._recording:
            return
        self._recording = False

        await self._stream_recorder.stop()
        if self.save_raw_danmaku:
            self._raw_danmaku_dumper.disable()
            self._raw_danmaku_receiver.stop()
        self._danmaku_dumper.disable()
        self._danmaku_receiver.stop()
        self._cover_downloader.disable()

        if self._stopped:
            logger.info('Recording Cancelled')
            await self._emit('recording_cancelled', self)
        else:
            logger.info('Recording Finished')
            await self._emit('recording_finished', self)

    async def _prepare(self) -> None:
        live_start_time = self._live.room_info.live_start_time
        self._danmaku_dumper.set_live_start_time(live_start_time)
        self._danmaku_dumper.clear_files()
        self._stream_recorder.clear_files()

    @aio_task_with_room_id
    async def _guard(self, timeout: float = 60) -> None:
        await asyncio.sleep(timeout)
        if not self._recording:
            return
        if self._stream_available:
            return
        logger.debug(
            f'Stream not available in {timeout} seconds, the event maybe lost.'
        )

        await self._live.update_info()
        if self._live.is_living():
            logger.debug('The live is living now')
            self._stream_available = True
            if self._stream_recorder.stopped:
                await self._stream_recorder.start()
        else:
            logger.debug('The live has ended before streaming')
            self._stream_available = False
            if not self._stream_recorder.stopped:
                await self.stop()

    def _print_waiting_message(self) -> None:
        logger.info('Waiting... until the live starts')

    def _print_live_info(self) -> None:
        room_info = self._live.room_info
        user_info = self._live.user_info

        if room_info.live_start_time > 0:
            live_start_time = str(datetime.fromtimestamp(room_info.live_start_time))
        else:
            live_start_time = 'NULL'

        msg = f"""
================================== User Info ==================================
user name        : {user_info.name}
gender           : {user_info.gender}
sign             : {user_info.sign}
uid              : {user_info.uid}
level            : {user_info.level}
---------------------------------- Room Info ----------------------------------
title            : {room_info.title}
cover            : {room_info.cover}
online           : {humanize.intcomma(room_info.online)}
live status      : {room_info.live_status.name}
live start time  : {live_start_time}
room id          : {room_info.room_id}
short room id    : {room_info.short_room_id or 'NULL'}
area id          : {room_info.area_id}
area name        : {room_info.area_name}
parent area id   : {room_info.parent_area_id}
parent area name : {room_info.parent_area_name}
tags             : {room_info.tags}
description      :
{room_info.description}
===============================================================================
"""
        logger.info(msg)

    def _print_changed_room_info(self, room_info: RoomInfo) -> None:
        msg = f"""
================================= Room Change =================================
title            : {room_info.title}
area id          ：{room_info.area_id}
area name        : {room_info.area_name}
parent area id   : {room_info.parent_area_id}
parent area name : {room_info.parent_area_name}
===============================================================================
"""
        logger.info(msg)



import asyncio
from typing import Any, Dict, Iterable, Literal

from ..flv.helpers import get_metadata as _get_metadata
from ..flv.helpers import get_extra_metadata as _get_extra_metadata


async def discard_files(
    paths: Iterable[str],
    log_level: Literal['INFO', 'DEBUG'] = 'INFO',
) -> None:
    for path in paths:
        await discard_file(path, log_level)


async def discard_file(
    path: str,
    log_level: Literal['INFO', 'DEBUG'] = 'INFO',
) -> None:
    from ..disk_space import delete_file
    await delete_file(path, log_level)


async def get_metadata(flv_path: str) -> Dict[str, Any]:
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, _get_metadata, flv_path)


async def get_extra_metadata(flv_path: str) -> Dict[str, Any]:
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, _get_extra_metadata, flv_path)



import logging
import math
from typing import Callable, Optional

from reactivex import Observable, abc

from ..common import (
    is_audio_tag,
    is_metadata_tag,
    is_script_tag,
    is_video_tag,
    parse_metadata,
)
from ..models import AudioTag, FlvHeader, FlvTag, ScriptTag, VideoTag
from .typing import FLVStream, FLVStreamItem

__all__ = ('fix',)

logger = logging.getLogger(__name__)


def fix() -> Callable[[FLVStream], FLVStream]:
    def _fix(source: FLVStream) -> FLVStream:
        """Fix broken timestamps of the FLV tags."""

        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            delta: int = 0
            last_tag: Optional[FlvTag] = None
            last_audio_tag: Optional[AudioTag] = None
            last_video_tag: Optional[VideoTag] = None
            frame_rate = 30.0
            video_frame_interval = math.ceil(1000 / frame_rate)
            sound_sample_interval = math.ceil(1000 / 44)

            def reset() -> None:
                nonlocal delta, last_tag, last_audio_tag, last_video_tag
                nonlocal frame_rate, video_frame_interval, sound_sample_interval
                delta = 0
                last_tag = None
                last_audio_tag = None
                last_video_tag = None
                frame_rate = 30.0
                video_frame_interval = math.ceil(1000 / frame_rate)
                sound_sample_interval = math.ceil(1000 / 44)

            def update_parameters(tag: ScriptTag) -> None:
                nonlocal frame_rate, video_frame_interval
                metadata = parse_metadata(tag)
                fps = metadata.get('fps') or metadata.get('framerate')

                if not fps:
                    return

                frame_rate = fps
                video_frame_interval = math.ceil(1000 / frame_rate)

                logger.debug(
                    'frame rate: {}, video frame interval: {}'.format(
                        frame_rate, video_frame_interval
                    )
                )

            def update_last_tags(tag: FlvTag) -> None:
                nonlocal last_tag, last_audio_tag, last_video_tag
                last_tag = tag
                if is_audio_tag(tag):
                    last_audio_tag = tag
                elif is_video_tag(tag):
                    last_video_tag = tag

            def update_delta(tag: FlvTag) -> None:
                nonlocal delta
                assert last_tag is not None
                delta = last_tag.timestamp + delta - tag.timestamp + calc_interval(tag)

            def correct_ts(tag: FlvTag) -> FlvTag:
                if delta == 0:
                    return tag
                return tag.evolve(timestamp=tag.timestamp + delta)

            def calc_interval(tag: FlvTag) -> int:
                if is_audio_tag(tag):
                    return sound_sample_interval
                elif is_video_tag(tag):
                    return video_frame_interval
                else:
                    logger.warning(f'Unexpected tag type: {tag}')
                    return min(sound_sample_interval, video_frame_interval)

            def is_ts_rebounded(tag: FlvTag) -> bool:
                if is_audio_tag(tag):
                    if last_audio_tag is None:
                        return False
                    return tag.timestamp < last_audio_tag.timestamp
                elif is_video_tag(tag):
                    if last_video_tag is None:
                        return False
                    return tag.timestamp < last_video_tag.timestamp
                else:
                    return False

            def is_ts_incontinuous(tag: FlvTag) -> bool:
                if last_tag is None:
                    return False
                return tag.timestamp - last_tag.timestamp > max(
                    sound_sample_interval, video_frame_interval
                )

            def on_next(item: FLVStreamItem) -> None:
                if isinstance(item, FlvHeader):
                    reset()
                    observer.on_next(item)
                    return

                tag = item

                if is_script_tag(tag):
                    if is_metadata_tag(tag):
                        update_parameters(tag)
                    observer.on_next(tag)
                    return

                if is_ts_rebounded(tag):
                    update_delta(tag)
                    logger.warning(
                        f'Timestamp rebounded, updated delta: {delta}\n'
                        f'last audio tag: {last_audio_tag}\n'
                        f'last video tag: {last_video_tag}\n'
                        f'current tag: {tag}'
                    )
                elif is_ts_incontinuous(tag):
                    update_delta(tag)
                    logger.warning(
                        f'Timestamp incontinuous, updated delta: {delta}\n'
                        f'last tag: {last_tag}\n'
                        f'current tag: {tag}'
                    )

                update_last_tags(tag)
                tag = correct_ts(tag)
                observer.on_next(tag)

            return source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

        return Observable(subscribe)

    return _fix



from __future__ import annotations
from abc import ABC, abstractmethod
from enum import IntEnum
from typing import Any, Final, Optional, TypeVar, cast
from typing_extensions import TypeGuard

import attr

from ..utils.hash import cksum


class TagType(IntEnum):
    AUDIO = 8
    VIDEO = 9
    SCRIPT = 18


class FrameType(IntEnum):
    KEY_FRAME = 1
    INNER_FRAME = 2
    DISPOSABLE_INNER_FRAME = 3
    GENERATED_KEY_FRAME = 4
    VIDEO_INFO = 5


class CodecID(IntEnum):
    SORENSON_H263 = 2
    SCREEN_VIDEO = 3
    ON2_VP6 = 4
    ON2_VP6_WITH_ALPHA_CHANNEL = 5
    SCREEN_VIDEO_V2 = 6
    AVC = 7


class AVCPacketType(IntEnum):
    AVC_SEQUENCE_HEADER = 0
    AVC_NALU = 1
    AVC_END_OF_SEQENCE = 2


class SoundFormat(IntEnum):
    LINEAR_PCM_PLATFORM_ENDIAN = 0
    ADPCM = 1
    MP3 = 2
    LINEAR_PCM_LITTLE_ENDIAN = 3
    NELLYMOSER_16KHZ_MONO = 4
    NELLYMOSER_8KHZ_MONO = 5
    NELLYMOSER = 6
    G711_A_LAW_LOGARITHMIC_PCM = 7
    G711_MU_LAW_LOGARITHMIC_PCM = 8
    AAC = 10
    SPEEX = 11
    MP3_8KHZ = 14
    DEVICE_SPECIFIC_SOUND = 15


class SoundRate(IntEnum):
    F_5_5KHZ = 0
    F_11KHZ = 1
    F_22KHZ = 2
    F_44KHZ = 3


class SoundSize(IntEnum):
    SAMPLES_8BIT = 0
    SAMPLES_16BIT = 1


class SoundType(IntEnum):
    MONO = 0
    STEREO = 1


class AACPacketType(IntEnum):
    AAC_SEQUENCE_HEADER = 0
    AAC_RAW = 1


def non_negative_integer_validator(instance, attribute, value):  # type: ignore
    if not isinstance(value, int) or value < 0:
        raise ValueError(f"'{attribute}' has to be a non negative integer!")


@attr.s(auto_attribs=True, slots=True)
class FlvHeader:
    signature: str
    version: int
    type_flag: int
    data_offset: int = attr.ib(validator=[non_negative_integer_validator])

    def has_video(self) -> bool:
        return bool(self.type_flag & 0b0000_0001)

    def has_audio(self) -> bool:
        return bool(self.type_flag & 0b0000_0100)

    def set_video_flag(self, value: bool) -> None:
        if value:
            self.type_flag |= 0b0000_0001
        else:
            self.type_flag &= ~0b0000_0001

    def set_audio_flag(self, value: bool) -> None:
        if value:
            self.type_flag |= 0b0000_0100
        else:
            self.type_flag &= ~0b0000_0100

    def __len__(self) -> int:
        return self.size

    @property
    def size(self) -> int:
        return self.data_offset


@attr.s(auto_attribs=True, slots=True, frozen=True)
class FlvTagHeader:
    filtered: bool
    tag_type: TagType
    data_size: int = attr.ib(validator=[non_negative_integer_validator])
    timestamp: int
    stream_id: int


@attr.s(auto_attribs=True, slots=True, frozen=True)
class AudioTagHeader:
    sound_format: SoundFormat
    sound_rate: SoundRate
    sound_size: SoundSize
    sound_type: SoundType
    aac_packet_type: Optional[AACPacketType]


@attr.s(auto_attribs=True, slots=True, frozen=True)
class VideoTagHeader:
    frame_type: FrameType
    codec_id: CodecID
    avc_packet_type: Optional[AVCPacketType]
    composition_time: Optional[int]


BACK_POINTER_SIZE: Final[int] = 4
TAG_HEADER_SIZE: Final[int] = 11
AUDIO_TAG_HEADER_SIZE: Final[int] = 2
VIDEO_TAG_HEADER_SIZE: Final[int] = 5


_T = TypeVar('_T', bound='FlvTag')


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class FlvTag(ABC, FlvTagHeader):
    offset: int = attr.ib(validator=[non_negative_integer_validator])
    body: Optional[bytes] = attr.ib(default=None, repr=cksum)

    def __len__(self) -> int:
        return self.tag_size

    @property
    @abstractmethod
    def header_size(self) -> int:
        ...

    @property
    def tag_size(self) -> int:
        return TAG_HEADER_SIZE + self.data_size

    @property
    def body_offset(self) -> int:
        return self.offset + TAG_HEADER_SIZE + self.header_size

    @property
    def body_size(self) -> int:
        return self.data_size - self.header_size

    @property
    def tag_end_offset(self) -> int:
        return self.offset + self.tag_size

    @property
    def next_tag_offset(self) -> int:
        return self.tag_end_offset + BACK_POINTER_SIZE

    def is_audio_tag(self) -> TypeGuard[AudioTag]:
        return self.tag_type == TagType.AUDIO

    def is_video_tag(self) -> TypeGuard[VideoTag]:
        return self.tag_type == TagType.VIDEO

    def is_script_tag(self) -> TypeGuard[ScriptTag]:
        return self.tag_type == TagType.SCRIPT

    def is_the_same_as(self, another: FlvTag) -> bool:
        return (
            self.tag_type == another.tag_type and
            self.data_size == another.data_size and
            self.body == another.body
        )

    def evolve(self: _T, **changes: Any) -> _T:
        if 'body' in changes:
            body = cast(bytes, changes.get('body'))
            changes['data_size'] = self.header_size + len(body)
        return attr.evolve(self, **changes)


# it’s not possible to inherit from more than one class that has attributes in
# __slots__
# ref: https://www.attrs.org/en/stable/glossary.html
# ref: https://docs.python.org/3/reference/datamodel.html#notes-on-using-slots
@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class AudioTag(FlvTag):
    sound_format: SoundFormat
    sound_rate: SoundRate
    sound_size: SoundSize
    sound_type: SoundType
    aac_packet_type: AACPacketType

    @property
    def header_size(self) -> int:
        if self.sound_format != SoundFormat.AAC:
            return 1
        else:
            return 2

    def is_aac_format(self) -> bool:
        return self.sound_format == SoundFormat.AAC

    def is_aac_header(self) -> bool:
        return self.aac_packet_type == AACPacketType.AAC_SEQUENCE_HEADER

    def is_aac_raw(self) -> bool:
        return self.aac_packet_type == AACPacketType.AAC_RAW


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class VideoTag(FlvTag):
    frame_type: FrameType
    codec_id: CodecID
    avc_packet_type: AVCPacketType
    composition_time: int

    @property
    def header_size(self) -> int:
        if self.codec_id != CodecID.AVC:
            return 1
        else:
            return 5

    def is_avc_header(self) -> bool:
        return self.avc_packet_type == AVCPacketType.AVC_SEQUENCE_HEADER

    def is_avc_nalu(self) -> bool:
        return self.avc_packet_type == AVCPacketType.AVC_NALU

    def is_avc_end(self) -> bool:
        return self.avc_packet_type == AVCPacketType.AVC_END_OF_SEQENCE

    def is_keyframe(self) -> bool:
        return self.frame_type == FrameType.KEY_FRAME


@attr.s(auto_attribs=True, slots=True, frozen=True, kw_only=True)
class ScriptTag(FlvTag):
    @property
    def header_size(self) -> int:
        return 0



from __future__ import annotations

import logging
from typing import Optional

from reactivex import Observable, abc

from ..common import (
    is_audio_sequence_header,
    is_metadata_tag,
    is_video_nalu_keyframe,
    is_video_sequence_header,
)
from ..models import AudioTag, FlvHeader, FlvTag, ScriptTag, VideoTag
from ..utils import format_timestamp
from .correct import correct
from .typing import FLVStream, FLVStreamItem

__all__ = ('Cutter',)


logger = logging.getLogger(__name__)


class Cutter:
    def __init__(self, min_duration: int = 5_000) -> None:
        self._min_duration = min_duration  # milliseconds
        self._reset()

    def _reset(self) -> None:
        self._cutting: bool = False
        self._triggered: bool = False
        self._last_timestamp: int = 0
        self._last_flv_header: Optional[FlvHeader] = None
        self._last_metadata_tag: Optional[ScriptTag] = None
        self._last_audio_sequence_header: Optional[AudioTag] = None
        self._last_video_sequence_header: Optional[VideoTag] = None

    def is_cutting(self) -> bool:
        return self._cutting

    def can_cut_stream(self) -> bool:
        if self._triggered or self._cutting:
            return False
        return self._last_timestamp >= self._min_duration

    def cut_stream(self) -> bool:
        if self.can_cut_stream():
            self._triggered = True
            return True
        return False

    def __call__(self, source: FLVStream) -> FLVStream:
        return self._cut(source).pipe(correct())

    def _cut(self, source: FLVStream) -> FLVStream:
        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            def on_next(item: FLVStreamItem) -> None:
                if isinstance(item, FlvHeader):
                    self._reset()
                    self._update_flv_header(item)
                else:
                    self._update_meta_tags(item)
                    self._check_cut_point(item)
                    if self._cutting:
                        self._insert_header_and_tags(observer)
                        self._cutting = False
                        self._triggered = False
                observer.on_next(item)

            return source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

        return Observable(subscribe)

    def _update_flv_header(self, header: FlvHeader) -> None:
        self._last_flv_header = header

    def _update_meta_tags(self, tag: FlvTag) -> None:
        if is_metadata_tag(tag):
            self._last_metadata_tag = tag
        elif is_audio_sequence_header(tag):
            self._last_audio_sequence_header = tag
        elif is_video_sequence_header(tag):
            self._last_video_sequence_header = tag

    def _insert_header_and_tags(
        self, observer: abc.ObserverBase[FLVStreamItem]
    ) -> None:
        assert self._last_flv_header is not None
        observer.on_next(self._last_flv_header)
        if self._last_metadata_tag is not None:
            observer.on_next(self._last_metadata_tag)
        if self._last_video_sequence_header is not None:
            observer.on_next(self._last_video_sequence_header)
        if self._last_audio_sequence_header is not None:
            observer.on_next(self._last_audio_sequence_header)

    def _check_cut_point(self, tag: FlvTag) -> None:
        self._last_timestamp = tag.timestamp

        if not self._triggered:
            return

        if not is_video_nalu_keyframe(tag):
            return

        self._cutting = True
        logger.info(f'Cut stream at {format_timestamp(tag.timestamp)}')



from __future__ import annotations

import logging
from typing import Final, Optional, Set

import m3u8
import urllib3
from ordered_set import OrderedSet
from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable

__all__ = ('PlaylistResolver',)


logger = logging.getLogger(__name__)
logging.getLogger(urllib3.__name__).setLevel(logging.WARNING)


class PlaylistResolver:
    _MAX_LAST_SEG_URIS: Final[int] = 30

    def __call__(self, source: Observable[m3u8.M3U8]) -> Observable[m3u8.Segment]:
        return self._solve(source)

    def _solve(self, source: Observable[m3u8.M3U8]) -> Observable[m3u8.Segment]:
        def subscribe(
            observer: abc.ObserverBase[m3u8.Segment],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()
            last_seg_uris: OrderedSet[str] = OrderedSet()

            def on_next(playlist: m3u8.M3U8) -> None:
                curr_seg_uris: Set[str] = set()

                for seg in playlist.segments:
                    if disposed:
                        return
                    curr_seg_uris.add(seg.uri)
                    if seg.uri not in last_seg_uris:
                        observer.on_next(seg)
                        last_seg_uris.add(seg.uri)
                        if len(last_seg_uris) > self._MAX_LAST_SEG_URIS:
                            last_seg_uris.pop(0)

                if last_seg_uris and not curr_seg_uris.intersection(last_seg_uris):
                    logger.debug(
                        'Segments broken!\n'
                        f'Last segments uris: {last_seg_uris}\n'
                        f'Current segments uris: {curr_seg_uris}'
                    )

                if playlist.is_endlist:
                    logger.debug('Playlist ended')

            def dispose() -> None:
                nonlocal disposed
                disposed = True
                last_seg_uris.clear()

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)



from .helpers import (
    file_exists,
    create_file,
    danmaku_path,
    cover_path,
    raw_danmaku_path,
    extra_metadata_path,
    escape_path,
)


__all__ = (
    'file_exists',
    'create_file',
    'danmaku_path',
    'cover_path',
    'raw_danmaku_path',
    'extra_metadata_path',
    'escape_path',
)



import logging
from typing import Callable

from .concat import concat
from .defragment import defragment
from .fix import fix
from .split import split
from .typing import FLVStream

__all__ = ('process',)

logger = logging.getLogger(__name__)


def process() -> Callable[[FLVStream], FLVStream]:
    def _process(source: FLVStream) -> FLVStream:
        return source.pipe(defragment(), split(), fix(), concat())

    return _process



import logging
from typing import Callable, List, Optional

from reactivex import Observable, abc

from ..models import FlvHeader
from .typing import FLVStream, FLVStreamItem

__all__ = ('defragment',)

logger = logging.getLogger(__name__)


def defragment(min_tags: int = 10) -> Callable[[FLVStream], FLVStream]:
    def _defragment(source: FLVStream) -> FLVStream:
        """Discard fragmented FLV streams."""

        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            gathering: bool = False
            gathered_items: List[FLVStreamItem] = []

            def on_next(item: FLVStreamItem) -> None:
                nonlocal gathering

                if isinstance(item, FlvHeader):
                    if gathered_items:
                        logger.debug(
                            'Discarded {} items, total size: {}'.format(
                                len(gathered_items), sum(len(t) for t in gathered_items)
                            )
                        )
                        gathered_items.clear()
                    gathering = True
                    logger.debug('Gathering items...')

                if gathering:
                    gathered_items.append(item)
                    if len(gathered_items) > min_tags:
                        for item in gathered_items:
                            observer.on_next(item)
                        gathered_items.clear()
                        gathering = False
                        logger.debug('Not a fragmented stream, stopped the gathering')
                else:
                    observer.on_next(item)

            return source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

        return Observable(subscribe)

    return _defragment



import logging
import os
import re
from datetime import datetime
from typing import Tuple

from ..bili.live import Live
from ..path import escape_path
from ..utils.mixins import AsyncCooperationMixin

__all__ = ('PathProvider',)

logger = logging.getLogger(__name__)


class PathProvider(AsyncCooperationMixin):
    def __init__(self, live: Live, out_dir: str, path_template: str) -> None:
        super().__init__()
        self._live = live
        self.out_dir = out_dir
        self.path_template = path_template

    def __call__(self) -> Tuple[str, int]:
        ts = self._run_coroutine(self._live.get_timestamp())
        path = self._make_path(ts)
        return path, ts

    def _make_path(self, timestamp: int) -> str:
        date_time = datetime.fromtimestamp(timestamp)
        relpath = self.path_template.format(
            roomid=self._live.room_id,
            uname=escape_path(self._live.user_info.name),
            title=escape_path(self._live.room_info.title),
            area=escape_path(self._live.room_info.area_name),
            parent_area=escape_path(self._live.room_info.parent_area_name),
            year=date_time.year,
            month=str(date_time.month).rjust(2, '0'),
            day=str(date_time.day).rjust(2, '0'),
            hour=str(date_time.hour).rjust(2, '0'),
            minute=str(date_time.minute).rjust(2, '0'),
            second=str(date_time.second).rjust(2, '0'),
        )

        pathname = os.path.abspath(
            os.path.expanduser(os.path.join(self.out_dir, relpath) + '.flv')
        )
        os.makedirs(os.path.dirname(pathname), exist_ok=True)
        while os.path.exists(pathname):
            root, ext = os.path.splitext(pathname)
            m = re.search(r'_\((\d+)\)$', root)
            if m is None:
                root += '_(1)'
            else:
                root = re.sub(r'\(\d+\)$', f'({int(m.group(1)) + 1})', root)
            pathname = root + ext

        return pathname



from __future__ import annotations

import logging
from typing import Optional, Union

import attr
import m3u8
import requests
import urllib3
from m3u8.model import InitializationSection
from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable
from tenacity import retry, retry_if_exception_type, stop_after_delay, wait_exponential

from ...bili.live import Live

__all__ = ('SegmentFetcher', 'InitSectionData', 'SegmentData')


logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, slots=True, frozen=True)
class InitSectionData:
    payload: bytes

    def __len__(self) -> int:
        return len(self.payload)


@attr.s(auto_attribs=True, slots=True, frozen=True)
class SegmentData:
    payload: bytes

    def __len__(self) -> int:
        return len(self.payload)


class SegmentFetcher:
    def __init__(self, live: Live, session: requests.Session) -> None:
        self._live = live
        self._session = session

    def __call__(
        self, source: Observable[m3u8.Segment]
    ) -> Observable[Union[InitSectionData, SegmentData]]:
        return self._fetch(source)

    def _fetch(
        self, source: Observable[m3u8.Segment]
    ) -> Observable[Union[InitSectionData, SegmentData]]:
        def subscribe(
            observer: abc.ObserverBase[Union[InitSectionData, SegmentData]],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()
            init_section: Optional[InitializationSection] = None

            def on_next(seg: m3u8.Segment) -> None:
                nonlocal init_section
                url: str = ''
                try:
                    if getattr(seg, 'init_section', None) and (
                        not init_section or seg.init_section.uri != init_section.uri
                    ):
                        url = seg.init_section.absolute_uri
                        data = self._fetch_segment(url)
                        init_section = seg.init_section
                        observer.on_next(InitSectionData(payload=data))
                    url = seg.absolute_uri
                    data = self._fetch_segment(url)
                    observer.on_next(SegmentData(payload=data))
                except Exception as e:
                    logger.warning(f'Failed to fetch segment {url}: {repr(e)}')

            def dispose() -> None:
                nonlocal disposed
                disposed = True

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)

    @retry(
        reraise=True,
        retry=retry_if_exception_type(
            (
                requests.exceptions.Timeout,
                requests.exceptions.HTTPError,
                urllib3.exceptions.TimeoutError,
                urllib3.exceptions.ProtocolError,
            )
        ),
        wait=wait_exponential(multiplier=0.1, max=5),
        stop=stop_after_delay(60),
    )
    def _fetch_segment(self, url: str) -> bytes:
        with self._session.get(url, headers=self._live.headers, timeout=10) as response:
            response.raise_for_status()
            return response.content



import logging
from typing import Optional

from reactivex.scheduler import NewThreadScheduler

from ..bili.live import Live
from ..bili.typing import QualityNumber
from ..flv import operators as flv_ops
from .stream_recorder_impl import StreamRecorderImpl

__all__ = ('FLVStreamRecorderImpl',)


logger = logging.getLogger(__name__)


class FLVStreamRecorderImpl(StreamRecorderImpl):
    def __init__(
        self,
        live: Live,
        out_dir: str,
        path_template: str,
        *,
        quality_number: QualityNumber = 10000,
        buffer_size: Optional[int] = None,
        read_timeout: Optional[int] = None,
        disconnection_timeout: Optional[int] = None,
        filesize_limit: int = 0,
        duration_limit: int = 0,
    ) -> None:
        super().__init__(
            live=live,
            out_dir=out_dir,
            path_template=path_template,
            stream_format='flv',
            quality_number=quality_number,
            buffer_size=buffer_size,
            read_timeout=read_timeout,
            disconnection_timeout=disconnection_timeout,
            filesize_limit=filesize_limit,
            duration_limit=duration_limit,
        )

    def _run(self) -> None:
        self._subscription = (
            self._stream_param_holder.get_stream_params()  # type: ignore
            .pipe(
                self._stream_url_resolver,
                self._stream_fetcher,
                self._dl_statistics,
                self._stream_parser,
                self._request_exception_handler,
                self._connection_error_handler,
                flv_ops.process(),
                self._cutter,
                self._limiter,
                self._join_point_extractor,
                self._prober,
                self._injector,
                self._analyser,
                self._dumper,
                self._rec_statistics,
                self._progress_bar,
                self._exception_handler,
            )
            .subscribe(
                on_completed=self._on_completed,
                scheduler=NewThreadScheduler(self._thread_factory('StreamRecorder')),
            )
        )



from enum import IntEnum
from datetime import datetime
from collections import OrderedDict
from typing import Any, BinaryIO, Dict, Final, List, Tuple


from .struct_io import StructWriter, StructReader


__all__ = (
    'Undefined',
    'AMFReader',
    'AMFWriter',
)


class ScriptDataValueType(IntEnum):
    NUMBER = 0
    BOOLEAN = 1
    STRING = 2
    OBJECT = 3
    MOVIE_CLIP = 4
    NULL = 5
    UNDEFINED = 6
    REFERENCE = 7
    ECMA_ARRAY = 8
    OBJECT_END_MARKER = 9
    STRICT_ARRAY = 10
    DATE = 11
    LONG_STRING = 12


Undefined: Final[object] = object()


class AMFReader:
    def __init__(self, stream: BinaryIO) -> None:
        self._reader = StructReader(stream)

    def read_value(self) -> Any:
        value_type = self._read_value_type()
        if value_type == ScriptDataValueType.NUMBER:
            return self._read_number()
        elif value_type == ScriptDataValueType.BOOLEAN:
            return self._read_boolean()
        elif value_type == ScriptDataValueType.STRING:
            return self._read_string()
        elif value_type == ScriptDataValueType.OBJECT:
            return self._read_object()
        elif value_type == ScriptDataValueType.MOVIE_CLIP:  # reserved
            raise NotImplementedError(value_type)
        elif value_type == ScriptDataValueType.NULL:
            return None
        elif value_type == ScriptDataValueType.UNDEFINED:
            return Undefined
        elif value_type == ScriptDataValueType.REFERENCE:
            raise NotImplementedError(value_type)
        elif value_type == ScriptDataValueType.ECMA_ARRAY:
            return self._read_ecma_array()
        elif value_type == ScriptDataValueType.OBJECT_END_MARKER:
            return value_type
        elif value_type == ScriptDataValueType.STRICT_ARRAY:
            return self._read_strict_array()
        elif value_type == ScriptDataValueType.DATE:
            return self._read_date()
        elif value_type == ScriptDataValueType.LONG_STRING:
            return self._read_long_string()
        else:
            raise ValueError(value_type)

    def _read_value_type(self) -> ScriptDataValueType:
        return ScriptDataValueType(self._reader.read_ui8())

    def _read_number(self) -> float:
        return self._reader.read_f64()

    def _read_boolean(self) -> bool:
        return self._reader.read_ui8() != 0

    def _read_string(self) -> str:
        size = self._reader.read_ui16()
        if size == 0:
            return ''
        data = self._reader.read(size)
        return data.decode()

    def _read_long_string(self) -> str:
        size = self._reader.read_ui32()
        if size == 0:
            return ''
        data = self._reader.read(size)
        return data.decode()

    def _read_object_property(self) -> Tuple[str, Any]:
        key = self._read_string()
        value = self.read_value()
        return key, value

    def _read_object(self) -> Dict[str, Any]:
        result = {}
        while True:
            key, value = self._read_object_property()
            if key == '' and value == ScriptDataValueType.OBJECT_END_MARKER:
                break
            result[key] = value
        return result

    def _read_ecma_array(self) -> 'OrderedDict[str, Any]':
        count = self._reader.read_ui32()
        result = OrderedDict(self._read_object())
        assert len(result) == count
        return result

    def _read_strict_array(self) -> List[Any]:
        count = self._reader.read_ui32()
        result = list(self.read_value() for _ in range(count))
        assert len(result) == count
        return result

    def _read_date(self) -> datetime:
        # the number of milliseconds
        timestamp = self._reader.read_f64() / 1000
        # reserved, should be set to 0x0000
        time_zone = self._reader.read_si16()
        assert time_zone == 0x0000
        return datetime.fromtimestamp(timestamp)


class AMFWriter:
    def __init__(self, stream: BinaryIO) -> None:
        self._writer = StructWriter(stream)

    def write_value(self, value: Any) -> None:
        if value is None:
            self._write_value_type(ScriptDataValueType.NULL)
        elif value is Undefined:
            self._write_value_type(ScriptDataValueType.UNDEFINED)
        elif isinstance(value, bool):
            self._write_value_type(ScriptDataValueType.BOOLEAN)
            self._write_boolean(value)
        elif isinstance(value, float):
            self._write_value_type(ScriptDataValueType.NUMBER)
            self._write_number(value)
        elif isinstance(value, str):
            if len(value.encode()) > 65535:  # XXX
                self._write_value_type(ScriptDataValueType.LONG_STRING)
                self._write_long_string(value)
            else:
                self._write_value_type(ScriptDataValueType.STRING)
                self._write_string(value)
        elif isinstance(value, list):
            self._write_value_type(ScriptDataValueType.STRICT_ARRAY)
            self._write_strict_array(value)
        elif isinstance(value, OrderedDict):
            self._write_value_type(ScriptDataValueType.ECMA_ARRAY)
            self._write_ecma_array(value)
        elif isinstance(value, dict):
            self._write_value_type(ScriptDataValueType.OBJECT)
            self._write_object(value)
        elif isinstance(value, datetime):
            self._write_value_type(ScriptDataValueType.DATE)
            self._write_date(value)
        else:
            raise TypeError(type(value))

    def _write_value_type(self, value_type: ScriptDataValueType) -> None:
        self._writer.write_ui8(value_type.value)

    def _write_number(self, number: float) -> None:
        self._writer.write_f64(number)

    def _write_boolean(self, boolean: bool) -> None:
        self._writer.write_ui8(int(boolean))

    def _write_string(self, string: str) -> None:
        data = string.encode()
        size = len(data)
        self._writer.write_ui16(size)
        if size > 0:
            self._writer.write(data)

    def _write_long_string(self, string: str) -> None:
        data = string.encode()
        size = len(data)
        self._writer.write_ui32(size)
        if size > 0:
            self._writer.write(data)

    def _write_object_property(self, key: str, value: Any) -> None:
        self._write_string(key)
        self.write_value(value)

    def _finalize_write_object(self) -> None:
        self._write_string('')
        self._write_value_type(ScriptDataValueType.OBJECT_END_MARKER)

    def _write_object(self, obj: Dict[str, Any]) -> None:
        for key, value in obj.items():
            self._write_object_property(key, value)
        self._finalize_write_object()

    def _write_ecma_array(self, array: 'OrderedDict[str, Any]') -> None:
        self._writer.write_ui32(len(array))
        self._write_object(array)

    def _write_strict_array(self, array: List[Any]) -> None:
        self._writer.write_ui32(len(array))
        for item in array:
            self.write_value(item)

    def _write_date(self, date_time: datetime) -> None:
        # the number of milliseconds
        self._writer.write_f64(date_time.timestamp() * 1000)
        # reserved, should be set to 0x0000
        self._writer.write_si16(0x0000)



from __future__ import annotations

import logging
from typing import Any, Final, Optional

import attr
from reactivex import Observable, abc, create
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable
from reactivex.scheduler.currentthreadscheduler import CurrentThreadScheduler

from ..bili.typing import ApiPlatform, QualityNumber, StreamFormat

__all__ = ('StreamParamHolder',)


logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class StreamParams:
    stream_format: StreamFormat
    quality_number: QualityNumber
    api_platform: ApiPlatform
    use_alternative_stream: bool


class StreamParamHolder:
    def __init__(
        self,
        *,
        stream_format: StreamFormat = 'flv',
        quality_number: QualityNumber = 10000,
        api_platform: ApiPlatform = 'web',
        use_alternative_stream: bool = False,
    ) -> None:
        super().__init__()
        self._stream_format: Final = stream_format
        self._quality_number = quality_number
        self._real_quality_number: Optional[QualityNumber] = None
        self._api_platform: ApiPlatform = api_platform
        self._use_alternative_stream: bool = use_alternative_stream
        self._cancelled: bool = False

    def reset(self) -> None:
        self._real_quality_number = None
        self._api_platform = 'web'
        self._use_alternative_stream = False
        self._cancelled = False

    def cancel(self) -> None:
        self._cancelled = True

    @property
    def stream_format(self) -> StreamFormat:
        return self._stream_format

    @property
    def quality_number(self) -> QualityNumber:
        return self._quality_number

    @quality_number.setter
    def quality_number(self, value: QualityNumber) -> None:
        self._quality_number = value

    @property
    def real_quality_number(self) -> QualityNumber:
        return self._real_quality_number or self._quality_number

    @property
    def use_alternative_stream(self) -> bool:
        return self._use_alternative_stream

    @use_alternative_stream.setter
    def use_alternative_stream(self, value: bool) -> None:
        self._use_alternative_stream = value

    def fall_back_quality(self) -> None:
        self._real_quality_number = 10000

    def rotate_api_platform(self) -> None:
        if self._api_platform == 'android':
            self._api_platform = 'web'
        else:
            self._api_platform = 'android'

    def get_stream_params(self) -> Observable[StreamParams]:
        self.reset()

        def subscribe(
            observer: abc.ObserverBase[StreamParams],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            _scheduler = scheduler or CurrentThreadScheduler.singleton()

            disposed = False
            cancelable = SerialDisposable()

            def action(
                scheduler: abc.SchedulerBase, state: Optional[Any] = None
            ) -> None:
                if self._cancelled or disposed:
                    return

                params = StreamParams(
                    stream_format=self._stream_format,
                    quality_number=self._real_quality_number or self._quality_number,
                    api_platform=self._api_platform,
                    use_alternative_stream=self._use_alternative_stream,
                )
                observer.on_next(params)

            cancelable.disposable = _scheduler.schedule(action)

            def dispose() -> None:
                nonlocal disposed
                disposed = True

            return CompositeDisposable(cancelable, Disposable(dispose))

        return create(subscribe)



from __future__ import annotations
from typing import TYPE_CHECKING

from .event_center import EventCenter
from .models import (
    LiveBeganEvent,
    LiveBeganEventData,
    LiveEndedEvent,
    LiveEndedEventData,
    RoomChangeEvent,
    RoomChangeEventData,
    RecordingStartedEvent,
    RecordingStartedEventData,
    RecordingFinishedEvent,
    RecordingFinishedEventData,
    RecordingCancelledEvent,
    RecordingCancelledEventData,
    VideoFileCreatedEvent,
    VideoFileCreatedEventData,
    VideoFileCompletedEvent,
    VideoFileCompletedEventData,
    DanmakuFileCreatedEvent,
    DanmakuFileCreatedEventData,
    DanmakuFileCompletedEvent,
    DanmakuFileCompletedEventData,
    RawDanmakuFileCreatedEvent,
    RawDanmakuFileCreatedEventData,
    RawDanmakuFileCompletedEvent,
    RawDanmakuFileCompletedEventData,
    VideoPostprocessingCompletedEvent,
    VideoPostprocessingCompletedEventData,
    SpaceNoEnoughEvent,
    SpaceNoEnoughEventData,
)
from ..bili.live import Live
from ..bili.models import RoomInfo
from ..bili.live_monitor import LiveEventListener
from ..core.recorder import RecorderEventListener
from ..disk_space import SpaceEventListener
from ..postprocess import PostprocessorEventListener
if TYPE_CHECKING:
    from ..bili.live_monitor import LiveMonitor
    from ..core.recorder import Recorder
    from ..disk_space import SpaceMonitor, DiskUsage
    from ..postprocess import Postprocessor


__all__ = (
    'LiveEventSubmitter',
    'SpaceEventSubmitter',
    'PostprocessorEventSubmitter',
)

event_center = EventCenter.get_instance()


class LiveEventSubmitter(LiveEventListener):
    def __init__(self, live_monitor: LiveMonitor) -> None:
        super().__init__()
        live_monitor.add_listener(self)

    async def on_live_began(self, live: Live) -> None:
        data = LiveBeganEventData(live.user_info, live.room_info)
        event_center.submit(LiveBeganEvent.from_data(data))

    async def on_live_ended(self, live: Live) -> None:
        data = LiveEndedEventData(live.user_info, live.room_info)
        event_center.submit(LiveEndedEvent.from_data(data))

    async def on_room_changed(self, room_info: RoomInfo) -> None:
        data = RoomChangeEventData(room_info)
        event_center.submit(RoomChangeEvent.from_data(data))


class RecorderEventSubmitter(RecorderEventListener):
    def __init__(self, recorder: Recorder) -> None:
        super().__init__()
        recorder.add_listener(self)

    async def on_recording_started(self, recorder: Recorder) -> None:
        data = RecordingStartedEventData(recorder.live.room_info)
        event_center.submit(RecordingStartedEvent.from_data(data))

    async def on_recording_finished(self, recorder: Recorder) -> None:
        data = RecordingFinishedEventData(recorder.live.room_info)
        event_center.submit(RecordingFinishedEvent.from_data(data))

    async def on_recording_cancelled(self, recorder: Recorder) -> None:
        data = RecordingCancelledEventData(recorder.live.room_info)
        event_center.submit(RecordingCancelledEvent.from_data(data))

    async def on_video_file_created(
        self, recorder: Recorder, path: str
    ) -> None:
        data = VideoFileCreatedEventData(recorder.live.room_id, path)
        event_center.submit(VideoFileCreatedEvent.from_data(data))

    async def on_video_file_completed(
        self, recorder: Recorder, path: str
    ) -> None:
        data = VideoFileCompletedEventData(recorder.live.room_id, path)
        event_center.submit(VideoFileCompletedEvent.from_data(data))

    async def on_danmaku_file_created(
        self, recorder: Recorder, path: str
    ) -> None:
        data = DanmakuFileCreatedEventData(recorder.live.room_id, path)
        event_center.submit(DanmakuFileCreatedEvent.from_data(data))

    async def on_danmaku_file_completed(
        self, recorder: Recorder, path: str
    ) -> None:
        data = DanmakuFileCompletedEventData(recorder.live.room_id, path)
        event_center.submit(DanmakuFileCompletedEvent.from_data(data))

    async def on_raw_danmaku_file_created(
        self, recorder: Recorder, path: str
    ) -> None:
        data = RawDanmakuFileCreatedEventData(recorder.live.room_id, path)
        event_center.submit(RawDanmakuFileCreatedEvent.from_data(data))

    async def on_raw_danmaku_file_completed(
        self, recorder: Recorder, path: str
    ) -> None:
        data = RawDanmakuFileCompletedEventData(recorder.live.room_id, path)
        event_center.submit(RawDanmakuFileCompletedEvent.from_data(data))


class PostprocessorEventSubmitter(PostprocessorEventListener):
    def __init__(self, postprocessor: Postprocessor) -> None:
        super().__init__()
        postprocessor.add_listener(self)

    async def on_video_postprocessing_completed(
        self, postprocessor: Postprocessor, path: str
    ) -> None:
        data = VideoPostprocessingCompletedEventData(
            postprocessor.recorder.live.room_id, path
        )
        event_center.submit(VideoPostprocessingCompletedEvent.from_data(data))


class SpaceEventSubmitter(SpaceEventListener):
    def __init__(self, space_monitor: SpaceMonitor) -> None:
        super().__init__()
        space_monitor.add_listener(self)

    async def on_space_no_enough(
        self, path: str, threshold: int, disk_usage: DiskUsage
    ) -> None:
        data = SpaceNoEnoughEventData(path, threshold, disk_usage)
        event_center.submit(SpaceNoEnoughEvent.from_data(data))




import asyncio

from .exception_center import ExceptionCenter


__all__ = (
    'ExceptionSubmitter',
    'submit_exception',
    'exception_callback',
)


class ExceptionSubmitter:
    def __enter__(self):  # type: ignore
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):  # type: ignore
        if exc_val is not None:
            submit_exception(exc_val)
        return True


def submit_exception(exc: BaseException) -> None:
    ExceptionCenter.get_instance().submit(exc)


def exception_callback(future: asyncio.Future) -> None:  # type: ignore
    if not future.done() or future.cancelled():
        return
    if (exc := future.exception()):
        submit_exception(exc)



import logging


from .exception_center import ExceptionCenter
from ..utils.mixins import SwitchableMixin


logger = logging.getLogger(__name__)


__all__ = 'ExceptionHandler',


class ExceptionHandler(SwitchableMixin):
    def _do_enable(self) -> None:
        exceptions = ExceptionCenter.get_instance().exceptions
        self._subscription = exceptions.subscribe(self._handle_exception)
        logger.debug('Enabled Exception Handler')

    def _do_disable(self) -> None:
        self._subscription.dispose()
        logger.debug('Disabled Exception Handler')

    def _handle_exception(self, exc: BaseException) -> None:
        self._log_exception(exc)

    def _log_exception(self, exc: BaseException) -> None:
        exc_info = (type(exc), exc, exc.__traceback__)
        logger.critical(type(exc).__name__, exc_info=exc_info)



from .connection_error_handler import ConnectionErrorHandler
from .exception_handler import ExceptionHandler
from .hls_prober import HLSProber, StreamProfile
from .playlist_fetcher import PlaylistFetcher
from .playlist_resolver import PlaylistResolver
from .progress_bar import ProgressBar
from .request_exception_handler import RequestExceptionHandler
from .segment_fetcher import InitSectionData, SegmentData, SegmentFetcher
from .segment_remuxer import SegmentRemuxer
from .sized_statistics import SizedStatistics
from .stream_fetcher import StreamFetcher
from .stream_parser import StreamParser
from .stream_statistics import StreamStatistics
from .stream_url_resolver import StreamURLResolver

__all__ = (
    'ConnectionErrorHandler',
    'ExceptionHandler',
    'HLSProber',
    'InitSectionData',
    'PlaylistFetcher',
    'PlaylistResolver',
    'ProgressBar',
    'RequestExceptionHandler',
    'SegmentData',
    'SegmentFetcher',
    'SegmentRemuxer',
    'SizedStatistics',
    'StreamFetcher',
    'StreamParser',
    'StreamProfile',
    'StreamStatistics',
    'StreamURLResolver',
)



import re


def snake_case(string: str) -> str:
    return re.sub(
        r'([a-z])([A-Z])',
        lambda m: m.group(1) + '_' + m.group(2).lower(),
        string
    )


def camel_case(string: str) -> str:
    words = string.split('_')
    return ''.join(
        [words[0].casefold()] + [word.capitalize() for word in words[1:]]
    )



import os
import zlib
import hashlib
from typing import Final, Union

from tqdm import tqdm


__all__ = 'cksum', 'md5sum', 'sha1sum'


CHUNK_SIZE: Final[int] = 8192


def cksum(data_or_path: Union[bytes, str], nopbar: bool = True) -> str:
    """Compute CRC32 checksum.

    data_or_path: can be bytes of binary data or a file path.
    """
    if isinstance(data_or_path, bytes):
        value = zlib.crc32(data_or_path, 0)
        return hex(value & 0xffffffff)[2:]

    with open(data_or_path, 'rb') as f, tqdm(
        desc='Computing CRC32',
        total=os.path.getsize(data_or_path),
        unit='B', unit_scale=True, unit_divisor=1024,
        disable=nopbar,
    ) as pbar:
        value = 0

        while True:
            data = f.read(CHUNK_SIZE)

            if not data:
                break

            value = zlib.crc32(data, value)
            pbar.update(len(data))

    return hex(value & 0xffffffff)[2:]


def md5sum(data_or_path: Union[bytes, str], nopbar: bool = True) -> str:
    """Compute MD5 message digest.

    data_or_path: can be bytes of binary data or a file path.
    """
    if isinstance(data_or_path, bytes):
        return hashlib.md5(data_or_path).hexdigest()

    with open(data_or_path, 'rb') as f, tqdm(
        desc='Computing MD5',
        total=os.path.getsize(data_or_path),
        postfix=os.path.basename(data_or_path),
        unit='B', unit_scale=True, unit_divisor=1024,
        disable=nopbar,
    ) as pbar:
        m = hashlib.md5()

        while True:
            data = f.read(CHUNK_SIZE)

            if not data:
                break

            m.update(data)
            pbar.update(len(data))

        md5 = m.hexdigest()

    return md5


def sha1sum(data_or_path: Union[bytes, str], nopbar: bool = True) -> str:
    """Compute SHA1 message digest.

    data_or_path: can be bytes of binary data or a file path.
    """
    if isinstance(data_or_path, bytes):
        return hashlib.sha1(data_or_path).hexdigest()

    with open(data_or_path, 'rb') as f, tqdm(
        desc='Computing SHA1',
        total=os.path.getsize(data_or_path),
        postfix=os.path.basename(data_or_path),
        unit='B', unit_scale=True, unit_divisor=1024,
        disable=nopbar,
    ) as pbar:
        s = hashlib.sha1()

        while True:
            data = f.read(CHUNK_SIZE)

            if not data:
                break

            s.update(data)
            pbar.update(len(data))

        sha1 = s.hexdigest()

    return sha1



from __future__ import annotations
import io
import math
from typing import Final, List, Tuple

import attr
from bitarray import bitarray

from .struct_io import StructReader
from .bits_io import BitsReader
from .utils import OffsetRepositor


__all__ = (
    'AVCDecoderConfigurationRecord',
    'SequenceParameterSet',
    'PictureParameterSet',
    'AVCSequenceHeaderParser',
    'NalUnit',
    'NalUnitParser',
    'SequenceParameterSetData',
    'SequenceParameterSetRBSPParser',
    'extract_resolution',
)


# ISO/IEC 14496-15:2010(E)
# 5.2.4.1.1 Syntax
# 5.2.4.1.2 Semantics

@attr.s(auto_attribs=True, slots=True, frozen=True)
class AVCDecoderConfigurationRecord:
    configuration_Version: int
    avc_profile_indication: int
    profile_compatibility: int
    avc_level_indication: int
    length_size_minus_one: int
    num_of_sequence_parameter_sets: int
    sequence_parameter_sets: List[SequenceParameterSet]
    num_of_picture_parameter_sets: int
    picture_parameter_sets: List[PictureParameterSet]
    # extensions ignored


@attr.s(auto_attribs=True, slots=True, frozen=True)
class SequenceParameterSet:
    sequence_parameter_set_length: int
    sequence_parameter_set_nal_unit: bytes


@attr.s(auto_attribs=True, slots=True, frozen=True)
class PictureParameterSet:
    picture_parameter_set_length: int
    picture_parameter_set_nal_unit: bytes


class AVCSequenceHeaderParser:
    def parse(self, data: bytes) -> AVCDecoderConfigurationRecord:
        reader = StructReader(io.BytesIO(data))

        configuration_Version = reader.read_ui8()
        avc_profile_indication = reader.read_ui8()
        profile_compatibility = reader.read_ui8()
        avc_level_indication = reader.read_ui8()
        length_size_minus_one = reader.read_ui8() & ~0b1111_1100

        num_of_sequence_parameter_sets = reader.read_ui8() & ~0b1110_0000
        sequence_parameter_sets: List[SequenceParameterSet] = []

        for _ in range(num_of_sequence_parameter_sets):
            sequence_parameter_set_length = reader.read_ui16()
            sequence_parameter_set_nal_unit = reader.read(
                sequence_parameter_set_length
            )
            sequence_parameter_sets.append(
                SequenceParameterSet(
                    sequence_parameter_set_length,
                    sequence_parameter_set_nal_unit,
                )
            )

        num_of_picture_parameter_sets = reader.read_ui8()
        picture_parameter_sets: List[PictureParameterSet] = []

        for _ in range(num_of_picture_parameter_sets):
            picture_parameter_set_length = reader.read_ui16()
            picture_parameter_set_nal_unit = reader.read(
                picture_parameter_set_length
            )
            picture_parameter_sets.append(
                PictureParameterSet(
                    picture_parameter_set_length,
                    picture_parameter_set_nal_unit,
                )
            )

        # extensions ignored

        return AVCDecoderConfigurationRecord(
            configuration_Version,
            avc_profile_indication,
            profile_compatibility,
            avc_level_indication,
            length_size_minus_one,
            num_of_sequence_parameter_sets,
            sequence_parameter_sets,
            num_of_picture_parameter_sets,
            picture_parameter_sets,
        )


# ISO/IEC 14496-10:2020(E)
# 7.3.1 NAL unit syntax
# 7.4.1 NAL unit semantics

@attr.s(auto_attribs=True, slots=True, frozen=True)
class NalUnit:
    forbidden_zero_bit: int
    nal_ref_idc: int
    nal_unit_type: int
    # extensions ignored
    rbsp_bytes: bytes


class NalUnitParser:
    def parse(self, data: bytes) -> NalUnit:
        stream = io.BytesIO(data)
        reader = StructReader(stream)

        byte = reader.read_ui8()
        forbidden_zero_bit = byte >> 7
        nal_ref_idc = (byte >> 5) & 0b0000_0011
        nal_unit_type = byte & 0b0001_1111

        # extensions ignored
        if nal_unit_type in (14, 20, 21):
            raise NotImplementedError()
        nal_unit_header_bytes = 1

        i = nal_unit_header_bytes
        num_bytes_in_nal_unit = len(data)
        rbsp_bytes_io = io.BytesIO()

        while i < num_bytes_in_nal_unit:
            with OffsetRepositor(stream):
                try:
                    next_24_bits = reader.read(3)
                except EOFError:
                    next_24_bits = b'0'

            if i + 2 < num_bytes_in_nal_unit and next_24_bits == 0x000003:
                rbsp_bytes_io.write(reader.read(2))
                emulation_prevention_three_byte = reader.read(1)
                assert emulation_prevention_three_byte == 0x03
                i += 3
            else:
                rbsp_bytes_io.write(reader.read(1))
                i += 1

        rbsp_bytes = rbsp_bytes_io.getvalue()

        return NalUnit(
            forbidden_zero_bit,
            nal_ref_idc,
            nal_unit_type,
            rbsp_bytes,
        )


# ISO/IEC 14496-10:2020(E)
# 7.3.2.1.1 Sequence parameter set data syntax
# 7.4.2.1.1 Sequence parameter set data semantics

# Table 6-1 – SubWidthC, and SubHeightC values derived from
# chroma_format_idc and separate_colour_plane_flag
SUB_WIDTH_HEIGHT_MAPPING: Final = {
    1: (2, 2),
    2: (2, 1),
    3: (1, 1),
}


@attr.s(auto_attribs=True, slots=True, frozen=True)
class SequenceParameterSetData:
    profile_idc: int
    constraint_set0_flag: int
    constraint_set1_flag: int
    constraint_set2_flag: int
    constraint_set3_flag: int
    constraint_set4_flag: int
    constraint_set5_flag: int
    level_idc: int
    seq_parameter_set_id: int
    chroma_format_idc: int
    separate_colour_plane_flag: int
    bit_depth_luma_minus8: int
    bit_depth_chroma_minus8: int
    qpprime_y_zero_transform_bypass_flag: int
    seq_scaling_matrix_present_flag: int
    seq_scaling_list_present_flag: List[int]
    log2_max_frame_num_minus4: int
    pic_order_cnt_type: int
    log2_max_pic_order_cnt_lsb_minus4: int
    delta_pic_order_always_zero_flag: int
    offset_for_non_ref_pic: int
    offset_for_top_to_bottom_field: int
    num_ref_frames_in_pic_order_cnt_cycle: int
    offset_for_ref_frame: List[int]
    max_num_ref_frames: int
    gaps_in_frame_num_value_allowed_flag: int
    pic_width_in_mbs_minus1: int
    pic_height_in_map_units_minus1: int
    frame_mbs_only_flag: int
    mb_adaptive_frame_field_flag: int
    direct_8x8_inference_flag: int
    frame_cropping_flag: int
    frame_crop_left_offset: int
    frame_crop_right_offset: int
    frame_crop_top_offset: int
    frame_crop_bottom_offset: int
    vui_parameters_present_flag: int

    @property
    def chroma_array_type(self) -> int:
        if self.separate_colour_plane_flag == 0:
            return self.chroma_format_idc
        return 0

    @property
    def sub_width_c(self) -> int:
        assert (
            self.chroma_format_idc in (1, 2, 3) and
            self.separate_colour_plane_flag == 0
        ), 'SubWidthC undefined!'
        return SUB_WIDTH_HEIGHT_MAPPING[self.chroma_format_idc][0]

    @property
    def sub_height_c(self) -> int:
        assert (
            self.chroma_format_idc in (1, 2, 3) and
            self.separate_colour_plane_flag == 0
        ), 'SubHeightC undefined!'
        return SUB_WIDTH_HEIGHT_MAPPING[self.chroma_format_idc][1]

    @property
    def mb_width_c(self) -> int:
        if self.chroma_format_idc == 0 or self.separate_colour_plane_flag == 1:
            return 0
        return 16 // self.sub_width_c

    @property
    def mb_height_c(self) -> int:
        if self.chroma_format_idc == 0 or self.separate_colour_plane_flag == 1:
            return 0
        return 16 // self.sub_height_c

    @property
    def pic_width_in_mbs(self) -> int:
        return self.pic_width_in_mbs_minus1 + 1

    @property
    def pic_width_in_samples_l(self) -> int:
        return self.pic_width_in_mbs * 16

    @property
    def pic_width_in_samples_c(self) -> int:
        return self.pic_width_in_mbs * self.mb_width_c

    @property
    def pic_height_in_map_units(self) -> int:
        return self.pic_height_in_map_units_minus1 + 1

    @property
    def pic_size_in_map_units(self) -> int:
        return self.pic_width_in_mbs * self.pic_height_in_map_units

    @property
    def frame_height_in_mbs(self) -> int:
        return (2 - self.frame_mbs_only_flag) * self.pic_height_in_map_units

    @property
    def crop_unit_x(self) -> int:
        if self.chroma_array_type == 0:
            return 1
        return self.sub_width_c

    @property
    def crop_unit_y(self) -> int:
        if self.chroma_array_type == 0:
            return 2 - self.frame_mbs_only_flag
        return self.sub_height_c * (2 - self.frame_mbs_only_flag)

    @property
    def frame_width(self) -> int:
        x0 = self.crop_unit_x * self.frame_crop_left_offset
        x1 = self.pic_width_in_samples_l - \
            (self.crop_unit_x * self.frame_crop_right_offset + 1)
        return x1 - x0 + 1  # x1 inclusive

    @property
    def frame_height(self) -> int:
        y0 = self.crop_unit_y * self.frame_crop_top_offset
        y1 = 16 * self.frame_height_in_mbs - \
            (self.crop_unit_y * self.frame_crop_bottom_offset + 1)
        return y1 - y0 + 1  # y1 inclusive


class SequenceParameterSetRBSPParser:
    def parse(self, rbsp: bytes) -> SequenceParameterSetData:
        bits = bitarray()
        bits.frombytes(rbsp)
        bits_reader = BitsReader(bits)
        egc_reader = ExpGolombCodeReader(bits_reader)

        profile_idc = bits_reader.read_bits_as_int(8)
        constraint_set0_flag = bits_reader.read_bits_as_int(1)
        constraint_set1_flag = bits_reader.read_bits_as_int(1)
        constraint_set2_flag = bits_reader.read_bits_as_int(1)
        constraint_set3_flag = bits_reader.read_bits_as_int(1)
        constraint_set4_flag = bits_reader.read_bits_as_int(1)
        constraint_set5_flag = bits_reader.read_bits_as_int(1)
        reserved_zero_2bits = bits_reader.read_bits_as_int(2)
        assert reserved_zero_2bits == 0
        level_idc = bits_reader.read_bits_as_int(8)
        seq_parameter_set_id = egc_reader.read_ue()

        # 7.4.2.1.1 Sequence parameter set data semantics
        # When chroma_format_idc is not present, it shall be inferred to be
        # equal to 1 (4:2:0 chroma format).
        chroma_format_idc = 1
        # When separate_colour_plane_flag is not present, it shall be
        # inferred to be equal to 0.
        separate_colour_plane_flag = 0
        # When bit_depth_luma_minus8 is not present, it shall be inferred to be
        # equal to 0. bit_depth_luma_minus8 shall be in the range of 0 to 6,
        # inclusive.
        bit_depth_luma_minus8 = 0
        # When bit_depth_chroma_minus8 is not present, it shall be inferred to
        # be equal to 0. bit_depth_chroma_minus8 shall be in the range of 0 to
        # 6, inclusive.
        bit_depth_chroma_minus8 = 0
        # When qpprime_y_zero_transform_bypass_flag is not present, it shall be
        # inferred to be equal to 0.
        qpprime_y_zero_transform_bypass_flag = 0
        # When seq_scaling_matrix_present_flag is not present, it shall be
        # inferred to be equal to 0.
        seq_scaling_matrix_present_flag = 0
        seq_scaling_list_present_flag = []

        if profile_idc in (
            100, 110, 122, 244, 44, 83, 86, 118, 128, 138, 139, 134, 135
        ):
            chroma_format_idc = egc_reader.read_ue()
            if chroma_format_idc == 3:
                separate_colour_plane_flag = bits_reader.read_bits_as_int(1)
            bit_depth_luma_minus8 = egc_reader.read_ue()
            bit_depth_chroma_minus8 = egc_reader.read_ue()
            qpprime_y_zero_transform_bypass_flag = \
                bits_reader.read_bits_as_int(1)
            seq_scaling_matrix_present_flag = bits_reader.read_bits_as_int(1)
            if seq_scaling_matrix_present_flag:
                for i in range(8 if chroma_format_idc != 3 else 12):
                    flag = bits_reader.read_bits_as_int(1)
                    seq_scaling_list_present_flag.append(flag)
                    if not flag:
                        continue
                    self._scaling_list(egc_reader, 16 if i < 6 else 64)

        log2_max_frame_num_minus4 = egc_reader.read_ue()
        pic_order_cnt_type = egc_reader.read_ue()

        log2_max_pic_order_cnt_lsb_minus4 = 0
        delta_pic_order_always_zero_flag = 1
        offset_for_non_ref_pic = 0
        offset_for_top_to_bottom_field = 0
        num_ref_frames_in_pic_order_cnt_cycle = 0
        offset_for_ref_frame = []

        if pic_order_cnt_type == 0:
            log2_max_pic_order_cnt_lsb_minus4 = egc_reader.read_ue()
        elif pic_order_cnt_type == 1:
            delta_pic_order_always_zero_flag = bits_reader.read_bits_as_int(1)
            offset_for_non_ref_pic = egc_reader.read_se()
            offset_for_top_to_bottom_field = egc_reader.read_se()
            num_ref_frames_in_pic_order_cnt_cycle = egc_reader.read_ue()
            offset_for_ref_frame = [
                egc_reader.read_se()
                for _ in range(num_ref_frames_in_pic_order_cnt_cycle)
            ]

        max_num_ref_frames = egc_reader.read_ue()
        gaps_in_frame_num_value_allowed_flag = bits_reader.read_bits_as_int(1)
        pic_width_in_mbs_minus1 = egc_reader.read_ue()
        pic_height_in_map_units_minus1 = egc_reader.read_ue()
        frame_mbs_only_flag = bits_reader.read_bits_as_int(1)

        # When mb_adaptive_frame_field_flag is not present, it shall be
        # inferred to be equal to 0.
        mb_adaptive_frame_field_flag = 0

        if not frame_mbs_only_flag:
            mb_adaptive_frame_field_flag = bits_reader.read_bits_as_int(1)

        direct_8x8_inference_flag = bits_reader.read_bits_as_int(1)
        frame_cropping_flag = bits_reader.read_bits_as_int(1)

        # When frame_cropping_flag is equal to 0, the values of
        # frame_crop_left_offset, frame_crop_right_offset,
        # frame_crop_top_offset, and frame_crop_bottom_offset shall be
        # inferred to be equal to 0.
        frame_crop_left_offset = 0
        frame_crop_right_offset = 0
        frame_crop_top_offset = 0
        frame_crop_bottom_offset = 0

        if frame_cropping_flag:
            frame_crop_left_offset = egc_reader.read_ue()
            frame_crop_right_offset = egc_reader.read_ue()
            frame_crop_top_offset = egc_reader.read_ue()
            frame_crop_bottom_offset = egc_reader.read_ue()

        vui_parameters_present_flag = bits_reader.read_bits_as_int(1)

        if vui_parameters_present_flag:
            # vui_parameters ignored
            pass

        # rbsp_trailing_bits ignored

        return SequenceParameterSetData(
            profile_idc,
            constraint_set0_flag,
            constraint_set1_flag,
            constraint_set2_flag,
            constraint_set3_flag,
            constraint_set4_flag,
            constraint_set5_flag,
            level_idc,
            seq_parameter_set_id,
            chroma_format_idc,
            separate_colour_plane_flag,
            bit_depth_luma_minus8,
            bit_depth_chroma_minus8,
            qpprime_y_zero_transform_bypass_flag,
            seq_scaling_matrix_present_flag,
            seq_scaling_list_present_flag,
            log2_max_frame_num_minus4,
            pic_order_cnt_type,
            log2_max_pic_order_cnt_lsb_minus4,
            delta_pic_order_always_zero_flag,
            offset_for_non_ref_pic,
            offset_for_top_to_bottom_field,
            num_ref_frames_in_pic_order_cnt_cycle,
            offset_for_ref_frame,
            max_num_ref_frames,
            gaps_in_frame_num_value_allowed_flag,
            pic_width_in_mbs_minus1,
            pic_height_in_map_units_minus1,
            frame_mbs_only_flag,
            mb_adaptive_frame_field_flag,
            direct_8x8_inference_flag,
            frame_cropping_flag,
            frame_crop_left_offset,
            frame_crop_right_offset,
            frame_crop_top_offset,
            frame_crop_bottom_offset,
            vui_parameters_present_flag,
        )

    def _scaling_list(
        self, egc_reader: ExpGolombCodeReader, list_size: int
    ) -> None:
        # 7.3.2.1.1.1 Scaling list syntax
        # scalingList and useDefaultScalingMatrixFlag ignored
        last_scale = next_scale = 8
        for _ in range(list_size):
            if next_scale != 0:
                delta_scale = egc_reader.read_se()
                next_scale = (last_scale + delta_scale + 256) % 256
            if next_scale != 0:
                last_scale = next_scale


# ISO/IEC 14496-10:2020(E)
# 9.1 Parsing process for Exp-Golomb codes
# 9.1.1 Mapping process for signed Exp-Golomb codes

class ExpGolombCodeReader:
    def __init__(self, reader: BitsReader) -> None:
        self._reader = reader

    def read_ue(self) -> int:
        zero_bit = bitarray('0')
        leading_zero_bits = 0

        while True:
            bit = self._reader.read_bits(1)
            if bit == zero_bit:
                leading_zero_bits += 1
            else:
                break

        return (
            2 ** leading_zero_bits - 1 +
            self._reader.read_bits_as_int(leading_zero_bits)
        )

    def read_se(self) -> int:
        code_num = self.read_ue()
        value = -1 ** (code_num + 1) * math.ceil(code_num / 2)
        result = value if code_num % 2 == 0 else abs(value)
        return result


def extract_resolution(packet: bytes) -> Tuple[int, int]:
    """Extract resolution from AVCDecoderConfigurationRecord packet."""
    record = AVCSequenceHeaderParser().parse(packet)
    sps = record.sequence_parameter_sets[0]
    nal_unit = NalUnitParser().parse(sps.sequence_parameter_set_nal_unit)
    sps_data = SequenceParameterSetRBSPParser().parse(nal_unit.rbsp_bytes)
    return sps_data.frame_width, sps_data.frame_height



import asyncio
import json
import logging
from contextlib import suppress
from threading import Lock

import aiofiles
from aiofiles.threadpool.text import AsyncTextIOWrapper
from tenacity import AsyncRetrying, retry_if_not_exception_type, stop_after_attempt

from ..bili.live import Live
from ..event.event_emitter import EventEmitter, EventListener
from ..exception import exception_callback, submit_exception
from ..logging.room_id import aio_task_with_room_id
from ..path import raw_danmaku_path
from ..utils.mixins import SwitchableMixin
from .raw_danmaku_receiver import RawDanmakuReceiver
from .stream_recorder import StreamRecorder, StreamRecorderEventListener

__all__ = 'RawDanmakuDumper', 'RawDanmakuDumperEventListener'


logger = logging.getLogger(__name__)


class RawDanmakuDumperEventListener(EventListener):
    async def on_raw_danmaku_file_created(self, path: str) -> None:
        ...

    async def on_raw_danmaku_file_completed(self, path: str) -> None:
        ...


class RawDanmakuDumper(
    EventEmitter[RawDanmakuDumperEventListener],
    StreamRecorderEventListener,
    SwitchableMixin,
):
    def __init__(
        self,
        live: Live,
        stream_recorder: StreamRecorder,
        danmaku_receiver: RawDanmakuReceiver,
    ) -> None:
        super().__init__()
        self._live = live  # @aio_task_with_room_id
        self._stream_recorder = stream_recorder
        self._receiver = danmaku_receiver
        self._lock: Lock = Lock()

    def _do_enable(self) -> None:
        self._stream_recorder.add_listener(self)
        logger.debug('Enabled raw danmaku dumper')

    def _do_disable(self) -> None:
        self._stream_recorder.remove_listener(self)
        logger.debug('Disabled raw danmaku dumper')

    async def on_video_file_created(
        self, video_path: str, record_start_time: int
    ) -> None:
        with self._lock:
            self._path = raw_danmaku_path(video_path)
            self._start_dumping()

    async def on_video_file_completed(self, video_path: str) -> None:
        with self._lock:
            await self._stop_dumping()

    def _start_dumping(self) -> None:
        self._create_dump_task()

    async def _stop_dumping(self) -> None:
        await self._cancel_dump_task()

    def _create_dump_task(self) -> None:
        self._dump_task = asyncio.create_task(self._do_dump())
        self._dump_task.add_done_callback(exception_callback)

    async def _cancel_dump_task(self) -> None:
        self._dump_task.cancel()
        with suppress(asyncio.CancelledError):
            await self._dump_task

    @aio_task_with_room_id
    async def _do_dump(self) -> None:
        logger.debug('Started dumping raw danmaku')
        try:
            async with aiofiles.open(self._path, 'wt', encoding='utf8') as f:
                logger.info(f"Raw danmaku file created: '{self._path}'")
                await self._emit('raw_danmaku_file_created', self._path)

                async for attempt in AsyncRetrying(
                    retry=retry_if_not_exception_type((asyncio.CancelledError)),
                    stop=stop_after_attempt(3),
                ):
                    with attempt:
                        try:
                            await self._dumping_loop(f)
                        except Exception as e:
                            submit_exception(e)
                            raise
                while True:
                    danmu = await self._receiver.get_raw_danmaku()
                    json_string = json.dumps(danmu, ensure_ascii=False)
                    await f.write(json_string + '\n')
        finally:
            logger.info(f"Raw danmaku file completed: '{self._path}'")
            await self._emit('raw_danmaku_file_completed', self._path)
            logger.debug('Stopped dumping raw danmaku')

    async def _dumping_loop(self, file: AsyncTextIOWrapper) -> None:
        while True:
            danmu = await self._receiver.get_raw_danmaku()
            json_string = json.dumps(danmu, ensure_ascii=False)
            await file.write(json_string + '\n')



import json
import logging
from typing import Iterable, cast

import aiofiles

from ..flv.helpers import make_comment_for_joinpoints
from ..flv.operators import JoinPoint
from .helpers import get_extra_metadata, get_metadata

logger = logging.getLogger(__name__)


async def make_metadata_file(flv_path: str) -> str:
    path = flv_path + '.meta'
    async with aiofiles.open(path, 'wb') as f:
        content = await _make_metadata_content(flv_path)
        await f.write(content.encode(encoding='utf8'))
    return path


async def _make_metadata_content(flv_path: str) -> str:
    metadata = await get_metadata(flv_path)
    extra_metadata = await get_extra_metadata(flv_path)

    comment = cast(str, metadata.get('Comment', ''))
    chapters = ''

    if join_points := extra_metadata.get('joinpoints'):
        join_points = list(map(JoinPoint.from_metadata_value, join_points))
        comment += '\n\n' + make_comment_for_joinpoints(join_points)
        duration = int(cast(float, metadata['duration']) * 1000)
        chapters = _make_chapters(join_points, duration)

    comment = '\\\n'.join(comment.splitlines())

    # ref: https://ffmpeg.org/ffmpeg-formats.html#Metadata-1
    return f"""\
;FFMETADATA1
Title={metadata['Title']}
Artist={metadata['Artist']}
Date={metadata['Date']}
# Description may be truncated!
Description={json.dumps(metadata['description'], ensure_ascii=False)}
Comment={comment}

{chapters}
"""


def _make_chapters(join_points: Iterable[JoinPoint], duration: int) -> str:
    join_points = filter(lambda p: not p.seamless, join_points)
    timestamps = list(map(lambda p: p.timestamp, join_points))
    if not timestamps:
        return ''
    timestamps.insert(0, 0)
    timestamps.append(duration)

    result = ''
    for i in range(1, len(timestamps)):
        start = timestamps[i - 1]
        end = timestamps[i]
        if end < start:
            logger.warning(f'Chapter end time {end} before start {start}')
            end = start
        result += f"""\
[CHAPTER]
TIMEBASE=1/1000
START={start}
END={end}
title=segment \\#{i}
"""
    return result



from typing import Any, Dict, List

import attr
from pydantic import PositiveInt, conint
from fastapi import (
    APIRouter,
    status,
    Body,
    Depends,
    BackgroundTasks,
)

from ..dependencies import task_data_filter, TaskDataFilter
from ..schemas import ResponseMessage
from ..responses import (
    not_found_responses,
    forbidden_responses,
    confict_responses,
    accepted_responses,
    created_responses,
)
from ...exception import NotFoundError, ForbiddenError
from ...application import Application


app: Application = None  # type: ignore  # bypass flake8 F821

router = APIRouter(
    prefix='/api/v1/tasks',
    tags=['tasks'],
)


@router.get('/data')
async def get_task_data(
    page: PositiveInt = 1,
    size: conint(ge=10, lt=100) = 100,  # type: ignore
    filter: TaskDataFilter = Depends(task_data_filter)
) -> List[Dict[str, Any]]:
    start = (page - 1) * size
    stop = page * size

    task_data = []
    for index, data in enumerate(filter(app.get_all_task_data())):
        if index < start:
            continue
        if index >= stop:
            break
        task_data.append(attr.asdict(data))

    return task_data


@router.get(
    '/{room_id}/data',
    responses={**not_found_responses},
)
async def get_one_task_data(room_id: int) -> Dict[str, Any]:
    return attr.asdict(app.get_task_data(room_id))


@router.get(
    '/{room_id}/param',
    responses={**not_found_responses},
)
async def get_task_param(room_id: int) -> Dict[str, Any]:
    return attr.asdict(app.get_task_param(room_id))


@router.get(
    '/{room_id}/metadata',
    responses={**not_found_responses},
)
async def get_task_metadata(room_id: int) -> Dict[str, Any]:
    metadata = app.get_task_metadata(room_id)
    if not metadata:
        return {}
    return attr.asdict(metadata)


@router.get(
    '/{room_id}/profile',
    responses={**not_found_responses},
)
async def get_task_stream_profile(room_id: int) -> Dict[str, Any]:
    return app.get_task_stream_profile(room_id)


@router.get(
    '/{room_id}/videos',
    responses={**not_found_responses},
)
async def get_task_video_file_details(room_id: int) -> List[Dict[str, Any]]:
    return [attr.asdict(d) for d in app.get_task_video_file_details(room_id)]


@router.get(
    '/{room_id}/danmakus',
    responses={**not_found_responses},
)
async def get_task_danmaku_file_details(room_id: int) -> List[Dict[str, Any]]:
    return [attr.asdict(d) for d in app.get_task_danmaku_file_details(room_id)]


@router.post(
    '/info',
    response_model=ResponseMessage,
    responses={**not_found_responses},
)
async def update_all_task_infos() -> ResponseMessage:
    await app.update_all_task_infos()
    return ResponseMessage(message='All task infos have been updated')


@router.post(
    '/{room_id}/info',
    response_model=ResponseMessage,
    responses={**not_found_responses},
)
async def update_task_info(room_id: int) -> ResponseMessage:
    await app.update_task_info(room_id)
    return ResponseMessage(message='The task info has been updated')


@router.get(
    '/{room_id}/cut',
    response_model=ResponseMessage,
    responses={**not_found_responses},
)
async def can_cut_stream(room_id: int) -> ResponseMessage:
    if app.can_cut_stream(room_id):
        return ResponseMessage(
            message='The stream can been cut',
            data={'result': True},
        )
    else:
        return ResponseMessage(
            message='The stream cannot been cut',
            data={'result': False},
        )


@router.post(
    '/{room_id}/cut',
    response_model=ResponseMessage,
    status_code=status.HTTP_202_ACCEPTED,
    responses={**not_found_responses, **forbidden_responses},
)
async def cut_stream(room_id: int) -> ResponseMessage:
    if not app.cut_stream(room_id):
        raise ForbiddenError('The stream cannot been cut')
    return ResponseMessage(message='The stream cutting have been triggered')


@router.post(
    '/start',
    response_model=ResponseMessage,
    responses={**not_found_responses},
)
async def start_all_tasks() -> ResponseMessage:
    await app.start_all_tasks()
    return ResponseMessage(message='All tasks have been started')


@router.post(
    '/{room_id}/start',
    response_model=ResponseMessage,
    responses={**not_found_responses},
)
async def start_task(room_id: int) -> ResponseMessage:
    await app.start_task(room_id)
    return ResponseMessage(message='The task has been started')


@router.post(
    '/stop',
    response_model=ResponseMessage,
    status_code=status.HTTP_202_ACCEPTED,
    responses={**accepted_responses},
)
async def stop_all_tasks(
    background_tasks: BackgroundTasks,
    force: bool = Body(False),
    background: bool = Body(False),
) -> ResponseMessage:
    if background:
        background_tasks.add_task(app.stop_all_tasks, force)
        return ResponseMessage(message='Stopping all tasks on the background')

    await app.stop_all_tasks(force)
    return ResponseMessage(message='All tasks have been stopped')


@router.post(
    '/{room_id}/stop',
    response_model=ResponseMessage,
    status_code=status.HTTP_202_ACCEPTED,
    responses={**not_found_responses, **accepted_responses},
)
async def stop_task(
    background_tasks: BackgroundTasks,
    room_id: int,
    force: bool = Body(False),
    background: bool = Body(False),
) -> ResponseMessage:
    if not app.has_task(room_id):
        raise NotFoundError(f'No task for the room {room_id}')

    if background:
        background_tasks.add_task(app.stop_task, room_id, force)
        return ResponseMessage(message='Stopping the task on the background')

    await app.stop_task(room_id, force)
    return ResponseMessage(message='The task has been stopped')


@router.post(
    '/recorder/enable',
    response_model=ResponseMessage,
)
async def enable_all_task_recorders() -> ResponseMessage:
    await app.enable_all_task_recorders()
    return ResponseMessage(message='All task recorders have been enabled')


@router.post(
    '/{room_id}/recorder/enable',
    response_model=ResponseMessage,
    responses={**not_found_responses},
)
async def enable_task_recorder(room_id: int) -> ResponseMessage:
    await app.enable_task_recorder(room_id)
    return ResponseMessage(message='The task recorder has been enabled')


@router.post(
    '/recorder/disable',
    response_model=ResponseMessage,
    status_code=status.HTTP_202_ACCEPTED,
    responses={**accepted_responses},
)
async def disable_all_task_recorders(
    background_tasks: BackgroundTasks,
    force: bool = Body(False),
    background: bool = Body(False),
) -> ResponseMessage:
    if background:
        background_tasks.add_task(app.disable_all_task_recorders, force)
        return ResponseMessage(
            message='Disabling all task recorders on the background'
        )

    await app.disable_all_task_recorders(force)
    return ResponseMessage(message='All task recorders have been disabled')


@router.post(
    '/{room_id}/recorder/disable',
    response_model=ResponseMessage,
    status_code=status.HTTP_202_ACCEPTED,
    responses={**not_found_responses, **accepted_responses},
)
async def disable_task_recorder(
    background_tasks: BackgroundTasks,
    room_id: int,
    force: bool = Body(False),
    background: bool = Body(False),
) -> ResponseMessage:
    if not app.has_task(room_id):
        raise NotFoundError(f'No task for the room {room_id}')

    if background:
        background_tasks.add_task(app.disable_task_recorder, room_id, force)
        return ResponseMessage(
            message='Disabling the task recorder on the background'
        )

    await app.disable_task_recorder(room_id, force)
    return ResponseMessage(message='The task recorder has been disabled')


@router.post(
    '/{room_id}',
    response_model=ResponseMessage,
    status_code=status.HTTP_201_CREATED,
    responses={
        **created_responses, **confict_responses, **forbidden_responses
    },
)
async def add_task(room_id: int) -> ResponseMessage:
    """Add a task for a room.

    the room_id argument can be both the real room id or the short room id.

    only use the real room id for task management.
    the short room id should be considered as a shorthand
    only for adding tasks conveniently.
    """
    real_room_id = await app.add_task(room_id)
    return ResponseMessage(
        message='Successfully Added Task',
        data={'room_id': real_room_id},
    )


@router.delete(
    '',
    response_model=ResponseMessage,
)
async def remove_all_tasks() -> ResponseMessage:
    await app.remove_all_tasks()
    return ResponseMessage(message='All tasks have been removed')


@router.delete(
    '/{room_id}',
    response_model=ResponseMessage,
    responses={**not_found_responses},
)
async def remove_task(room_id: int) -> ResponseMessage:
    await app.remove_task(room_id)
    return ResponseMessage(message='The task has been removed')



import io
import logging
from typing import Callable, Optional

from reactivex import Observable, abc
from reactivex.disposable import CompositeDisposable, Disposable, SerialDisposable

from ..io import FlvReader
from .typing import FLVStream, FLVStreamItem

__all__ = ('parse',)

logger = logging.getLogger(__name__)


def parse(
    *,
    ignore_eof: bool = False,
    complete_on_eof: bool = False,
    ignore_value_error: bool = False,
    backup_timestamp: bool = False,
    restore_timestamp: bool = False,
) -> Callable[[Observable[io.RawIOBase]], FLVStream]:
    def _parse(source: Observable[io.RawIOBase]) -> FLVStream:
        def subscribe(
            observer: abc.ObserverBase[FLVStreamItem],
            scheduler: Optional[abc.SchedulerBase] = None,
        ) -> abc.DisposableBase:
            disposed = False
            subscription = SerialDisposable()

            def on_next(stream: io.RawIOBase) -> None:
                try:
                    try:
                        reader = FlvReader(
                            stream,
                            backup_timestamp=backup_timestamp,
                            restore_timestamp=restore_timestamp,
                        )
                        observer.on_next(reader.read_header())
                        while not disposed:
                            tag = reader.read_tag()
                            observer.on_next(tag)
                    finally:
                        stream.close()
                except EOFError as e:
                    if complete_on_eof:
                        observer.on_completed()
                    else:
                        if not ignore_eof:
                            observer.on_error(e)
                except ValueError as e:
                    if ignore_value_error:
                        logger.debug(f'Error occurred while parsing stream: {repr(e)}')
                    else:
                        observer.on_error(e)
                except Exception as e:
                    observer.on_error(e)
                else:
                    observer.on_completed()

            subscription.disposable = source.subscribe(
                on_next, observer.on_error, observer.on_completed, scheduler=scheduler
            )

            def dispose() -> None:
                nonlocal disposed
                disposed = True

            return CompositeDisposable(subscription, Disposable(dispose))

        return Observable(subscribe)

    return _parse



from .postprocessor import Postprocessor, PostprocessorEventListener
from .models import PostprocessorStatus, DeleteStrategy


__all__ = (
    'Postprocessor',
    'PostprocessorEventListener',
    'PostprocessorStatus',
    'DeleteStrategy',
)



import logging
import asyncio
import json

from fastapi import (
    APIRouter,
    WebSocket,
    WebSocketDisconnect,
)
from websockets.exceptions import ConnectionClosed

from ...event import EventCenter
from ...event.typing import Event
from ...exception import ExceptionCenter, format_exception
from ...application import Application


logger = logging.getLogger(__name__)
logging.getLogger('websockets').setLevel(logging.WARNING)

app: Application = None  # type: ignore  # bypass flake8 F821

router = APIRouter(
    tags=['websockets'],
)


@router.websocket('/ws/v1/events')
async def receive_events(websocket: WebSocket) -> None:
    await websocket.accept()
    logger.debug('Events websocket accepted')

    future = asyncio.Future()  # type: ignore

    async def send_event(event: Event) -> None:
        try:
            text = json.dumps(event.asdict(), ensure_ascii=False)
            await websocket.send_text(text)
        except (WebSocketDisconnect, ConnectionClosed) as e:
            logger.debug(f'Events websocket closed: {repr(e)}')
            subscription.dispose()
            future.set_result(None)
        except Exception as e:
            logger.error(f'Error occurred on events websocket: {repr(e)}')
            subscription.dispose()
            future.set_exception(e)

    def on_event(event: Event) -> None:
        asyncio.create_task(send_event(event))

    subscription = EventCenter.get_instance().events.subscribe(on_event)

    await future


@router.websocket('/ws/v1/exceptions')
async def receive_exception(websocket: WebSocket) -> None:
    await websocket.accept()
    logger.debug('Exceptions websocket accepted')

    future = asyncio.Future()  # type: ignore

    async def send_exception(exc: BaseException) -> None:
        try:
            await websocket.send_text(format_exception(exc))
        except (WebSocketDisconnect, ConnectionClosed) as e:
            logger.debug(f'Exceptions websocket closed: {repr(e)}')
            subscription.dispose()
            future.set_result(None)
        except Exception as e:
            logger.error(f'Error occurred on exceptions websocket: {repr(e)}')
            subscription.dispose()
            future.set_exception(e)

    def on_exception(exc: BaseException) -> None:
        asyncio.create_task(send_exception(exc))

    subscription = ExceptionCenter.get_instance().exceptions \
        .subscribe(on_exception)

    await future



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Optional, cast

from ..exception import NotFoundError
from ..logging import configure_logger
from ..notification import (
    EmailService,
    Notifier,
    Pushdeer,
    Pushplus,
    Serverchan,
    Telegram,
)
from ..webhook import WebHook
from .helpers import shadow_settings, update_settings
from .models import (
    DanmakuOptions,
    HeaderOptions,
    MessageTemplateSettings,
    NotificationSettings,
    NotifierSettings,
    OutputOptions,
    PostprocessingOptions,
    RecorderOptions,
    Settings,
    SettingsIn,
    SettingsOut,
    TaskOptions,
    TaskSettings,
)
from .typing import KeySetOfSettings

if TYPE_CHECKING:
    from ..application import Application


class SettingsManager:
    def __init__(self, app: Application, settings: Settings) -> None:
        self._app = app
        self._settings = settings

    def get_settings(
        self,
        include: Optional[KeySetOfSettings] = None,
        exclude: Optional[KeySetOfSettings] = None,
    ) -> SettingsOut:
        return SettingsOut(**self._settings.dict(include=include, exclude=exclude))

    async def change_settings(self, settings: SettingsIn) -> SettingsOut:
        changed = False

        for name in settings.__fields_set__:
            src_sub_settings = getattr(settings, name)
            dst_sub_settings = getattr(self._settings, name)

            if src_sub_settings == dst_sub_settings:
                continue

            if isinstance(src_sub_settings, list):
                assert isinstance(dst_sub_settings, list)
                setattr(self._settings, name, src_sub_settings)
            else:
                update_settings(src_sub_settings, dst_sub_settings)
            changed = True

            func = getattr(self, f'apply_{name}_settings')
            if asyncio.iscoroutinefunction(func):
                await func()
            else:
                func()

        if changed:
            await self.dump_settings()

        return self.get_settings(cast(KeySetOfSettings, settings.__fields_set__))

    def get_task_options(self, room_id: int) -> TaskOptions:
        if settings := self.find_task_settings(room_id):
            return TaskOptions.from_settings(settings)
        raise NotFoundError(f'task settings of room {room_id} not found')

    async def change_task_options(
        self, room_id: int, options: TaskOptions
    ) -> TaskOptions:
        settings = self.find_task_settings(room_id)
        assert settings is not None

        changed = False

        for name in options.__fields_set__:
            src_opts = getattr(options, name)
            dst_opts = getattr(settings, name)

            if src_opts == dst_opts:
                continue

            update_settings(src_opts, dst_opts)
            changed = True

            func = getattr(self, f'apply_task_{name}_settings')
            if asyncio.iscoroutinefunction(func):
                await func(room_id, dst_opts)
            else:
                func(room_id, dst_opts)

        if changed:
            await self.dump_settings()

        return TaskOptions.from_settings(settings)

    async def dump_settings(self) -> None:
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self._settings.dump)

    def has_task_settings(self, room_id: int) -> bool:
        return self.find_task_settings(room_id) is not None

    def find_task_settings(self, room_id: int) -> Optional[TaskSettings]:
        for settings in self._settings.tasks:
            if settings.room_id == room_id:
                return settings
        return None

    async def add_task_settings(self, room_id: int) -> TaskSettings:
        settings = TaskSettings(room_id=room_id)
        self._settings.tasks = [*self._settings.tasks, settings]
        await self.dump_settings()
        return settings.copy(deep=True)

    async def remove_task_settings(self, room_id: int) -> None:
        settings = self.find_task_settings(room_id)
        if settings is None:
            raise NotFoundError(f"The room {room_id} is not existed")
        self._settings.tasks.remove(settings)
        await self.dump_settings()

    async def remove_all_task_settings(self) -> None:
        self._settings.tasks.clear()
        await self.dump_settings()

    async def mark_task_enabled(self, room_id: int) -> None:
        settings = self.find_task_settings(room_id)
        assert settings is not None
        settings.enable_monitor = True
        settings.enable_recorder = True
        await self.dump_settings()

    async def mark_task_disabled(self, room_id: int) -> None:
        settings = self.find_task_settings(room_id)
        assert settings is not None
        settings.enable_monitor = False
        settings.enable_recorder = False
        await self.dump_settings()

    async def mark_all_tasks_enabled(self) -> None:
        for settings in self._settings.tasks:
            settings.enable_monitor = True
            settings.enable_recorder = True
        await self.dump_settings()

    async def mark_all_tasks_disabled(self) -> None:
        for settings in self._settings.tasks:
            settings.enable_monitor = False
            settings.enable_recorder = False
        await self.dump_settings()

    async def mark_task_monitor_enabled(self, room_id: int) -> None:
        settings = self.find_task_settings(room_id)
        assert settings is not None
        settings.enable_monitor = True
        await self.dump_settings()

    async def mark_task_monitor_disabled(self, room_id: int) -> None:
        settings = self.find_task_settings(room_id)
        assert settings is not None
        settings.enable_monitor = False
        await self.dump_settings()

    async def mark_all_task_monitors_enabled(self) -> None:
        for settings in self._settings.tasks:
            settings.enable_monitor = True
        await self.dump_settings()

    async def mark_all_task_monitors_disabled(self) -> None:
        for settings in self._settings.tasks:
            settings.enable_monitor = False
        await self.dump_settings()

    async def mark_task_recorder_enabled(self, room_id: int) -> None:
        settings = self.find_task_settings(room_id)
        assert settings is not None
        settings.enable_recorder = True
        await self.dump_settings()

    async def mark_task_recorder_disabled(self, room_id: int) -> None:
        settings = self.find_task_settings(room_id)
        assert settings is not None
        settings.enable_recorder = False
        await self.dump_settings()

    async def mark_all_task_recorders_enabled(self) -> None:
        for settings in self._settings.tasks:
            settings.enable_recorder = True
        await self.dump_settings()

    async def mark_all_task_recorders_disabled(self) -> None:
        for settings in self._settings.tasks:
            settings.enable_recorder = False
        await self.dump_settings()

    async def apply_task_header_settings(
        self, room_id: int, options: HeaderOptions, *, update_session: bool = True
    ) -> None:
        final_settings = self._settings.header.copy()
        shadow_settings(options, final_settings)
        await self._app._task_manager.apply_task_header_settings(
            room_id, final_settings, update_session=update_session
        )

    def apply_task_danmaku_settings(
        self, room_id: int, options: DanmakuOptions
    ) -> None:
        final_settings = self._settings.danmaku.copy()
        shadow_settings(options, final_settings)
        self._app._task_manager.apply_task_danmaku_settings(room_id, final_settings)

    def apply_task_recorder_settings(
        self, room_id: int, options: RecorderOptions
    ) -> None:
        final_settings = self._settings.recorder.copy()
        shadow_settings(options, final_settings)
        self._app._task_manager.apply_task_recorder_settings(room_id, final_settings)

    def apply_task_output_settings(self, room_id: int, options: OutputOptions) -> None:
        final_settings = self._settings.output.copy()
        shadow_settings(options, final_settings)
        self._app._task_manager.apply_task_output_settings(room_id, final_settings)

    def apply_task_postprocessing_settings(
        self, room_id: int, options: PostprocessingOptions
    ) -> None:
        final_settings = self._settings.postprocessing.copy()
        shadow_settings(options, final_settings)
        self._app._task_manager.apply_task_postprocessing_settings(
            room_id, final_settings
        )

    async def apply_output_settings(self) -> None:
        for settings in self._settings.tasks:
            self.apply_task_output_settings(settings.room_id, settings.output)

        out_dir = self._settings.output.out_dir
        self._app._out_dir = out_dir
        self._app._space_monitor.path = out_dir
        self._app._space_reclaimer.path = out_dir

    def apply_logging_settings(self) -> None:
        configure_logger(
            log_dir=self._settings.logging.log_dir,
            console_log_level=self._settings.logging.console_log_level,
            max_bytes=self._settings.logging.max_bytes,
            backup_count=self._settings.logging.backup_count,
        )

    async def apply_header_settings(self) -> None:
        for settings in self._settings.tasks:
            await self.apply_task_header_settings(settings.room_id, settings.header)

    def apply_danmaku_settings(self) -> None:
        for settings in self._settings.tasks:
            self.apply_task_danmaku_settings(settings.room_id, settings.danmaku)

    def apply_recorder_settings(self) -> None:
        for settings in self._settings.tasks:
            self.apply_task_recorder_settings(settings.room_id, settings.recorder)

    def apply_postprocessing_settings(self) -> None:
        for settings in self._settings.tasks:
            self.apply_task_postprocessing_settings(
                settings.room_id, settings.postprocessing
            )

    def apply_space_settings(self) -> None:
        self.apply_space_monitor_settings()
        self.apply_space_reclaimer_settings()

    def apply_space_monitor_settings(self) -> None:
        settings = self._settings.space
        self._app._space_monitor.check_interval = settings.check_interval
        self._app._space_monitor.space_threshold = settings.space_threshold

    def apply_space_reclaimer_settings(self) -> None:
        settings = self._settings.space
        self._app._space_reclaimer.recycle_records = settings.recycle_records

    def apply_email_notification_settings(self) -> None:
        notifier = self._app._email_notifier
        settings = self._settings.email_notification
        self._apply_email_settings(notifier.provider)
        self._apply_notifier_settings(notifier, settings)
        self._apply_notification_settings(notifier, settings)
        self._apply_message_template_settings(notifier, settings)

    def apply_serverchan_notification_settings(self) -> None:
        notifier = self._app._serverchan_notifier
        settings = self._settings.serverchan_notification
        self._apply_serverchan_settings(notifier.provider)
        self._apply_notifier_settings(notifier, settings)
        self._apply_notification_settings(notifier, settings)
        self._apply_message_template_settings(notifier, settings)

    def apply_pushdeer_notification_settings(self) -> None:
        notifier = self._app._pushdeer_notifier
        settings = self._settings.pushdeer_notification
        self._apply_pushdeer_settings(notifier.provider)
        self._apply_notifier_settings(notifier, settings)
        self._apply_notification_settings(notifier, settings)
        self._apply_message_template_settings(notifier, settings)

    def apply_pushplus_notification_settings(self) -> None:
        notifier = self._app._pushplus_notifier
        settings = self._settings.pushplus_notification
        self._apply_pushplus_settings(notifier.provider)
        self._apply_notifier_settings(notifier, settings)
        self._apply_notification_settings(notifier, settings)
        self._apply_message_template_settings(notifier, settings)

    def apply_telegram_notification_settings(self) -> None:
        notifier = self._app._telegram_notifier
        settings = self._settings.telegram_notification
        self._apply_telegram_settings(notifier.provider)
        self._apply_notifier_settings(notifier, settings)
        self._apply_notification_settings(notifier, settings)
        self._apply_message_template_settings(notifier, settings)

    def apply_webhooks_settings(self) -> None:
        webhooks = [WebHook.from_settings(s) for s in self._settings.webhooks]
        self._app._webhook_emitter.webhooks = webhooks

    def _apply_email_settings(self, email_service: EmailService) -> None:
        email_service.src_addr = self._settings.email_notification.src_addr
        email_service.dst_addr = self._settings.email_notification.dst_addr
        email_service.auth_code = self._settings.email_notification.auth_code
        email_service.smtp_host = self._settings.email_notification.smtp_host
        email_service.smtp_port = self._settings.email_notification.smtp_port

    def _apply_serverchan_settings(self, serverchan: Serverchan) -> None:
        serverchan.sendkey = self._settings.serverchan_notification.sendkey

    def _apply_pushdeer_settings(self, pushdeer: Pushdeer) -> None:
        pushdeer.server = self._settings.pushdeer_notification.server
        pushdeer.pushkey = self._settings.pushdeer_notification.pushkey

    def _apply_pushplus_settings(self, pushplus: Pushplus) -> None:
        pushplus.token = self._settings.pushplus_notification.token
        pushplus.topic = self._settings.pushplus_notification.topic

    def _apply_telegram_settings(self, telegram: Telegram) -> None:
        telegram.token = self._settings.telegram_notification.token
        telegram.chatid = self._settings.telegram_notification.chatid

    def _apply_notifier_settings(
        self, notifier: Notifier, settings: NotifierSettings
    ) -> None:
        if settings.enabled:
            notifier.enable()
        else:
            notifier.disable()

    def _apply_notification_settings(
        self, notifier: Notifier, settings: NotificationSettings
    ) -> None:
        notifier.notify_began = settings.notify_began
        notifier.notify_ended = settings.notify_ended
        notifier.notify_error = settings.notify_error
        notifier.notify_space = settings.notify_space

    def _apply_message_template_settings(
        self, notifier: Notifier, settings: MessageTemplateSettings
    ) -> None:
        notifier.began_message_type = settings.began_message_type
        notifier.began_message_title = settings.began_message_title
        notifier.began_message_content = settings.began_message_content
        notifier.ended_message_type = settings.ended_message_type
        notifier.ended_message_title = settings.ended_message_title
        notifier.ended_message_content = settings.ended_message_content
        notifier.space_message_type = settings.space_message_type
        notifier.space_message_title = settings.space_message_title
        notifier.space_message_content = settings.space_message_content
        notifier.error_message_type = settings.error_message_type
        notifier.error_message_title = settings.error_message_title
        notifier.error_message_content = settings.error_message_content
